#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Shared Response Model (SRM)

The implementations are based on the following publications:

.. [Chen2015] "A Reduced-Dimension fMRI Shared Response Model",
   P.-H. Chen, J. Chen, Y. Yeshurun-Dishon, U. Hasson, J. Haxby, P. Ramadge
   Advances in Neural Information Processing Systems (NIPS), 2015.
   http://papers.nips.cc/paper/5855-a-reduced-dimension-fmri-shared-response-model

.. [Anderson2016] "Enabling Factor Analysis on Thousand-Subject Neuroimaging
   Datasets",
   Michael J. Anderson, Mihai Capotă, Javier S. Turek, Xia Zhu, Theodore L.
   Willke, Yida Wang, Po-Hsuan Chen, Jeremy R. Manning, Peter J. Ramadge,
   Kenneth A. Norman,
   IEEE International Conference on Big Data, 2016.
   https://doi.org/10.1109/BigData.2016.7840719
"""

# Authors: Po-Hsuan Chen (Princeton Neuroscience Institute) and Javier Turek
# (Intel Labs), 2015

import logging

import numpy as np
import scipy
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import assert_all_finite
from sklearn.exceptions import NotFittedError
from mpi4py import MPI
import sys

__all__ = [
    "DetSRM",
    "SRM",
]

logger = logging.getLogger(__name__)


def _init_w_transforms(data, features, random_states, comm=MPI.COMM_SELF):
    """Initialize the mappings (Wi) for the SRM with random orthogonal
    matrices.

    Parameters
    ----------

    data : list of 2D arrays, element i has shape=[voxels_i, samples]
        Each element in the list contains the fMRI data of one subject.

    features : int
        The number of features in the model.

    random_states : list of `RandomState`s
        One `RandomState` instance per subject.

    comm : mpi4py.MPI.Intracomm
        The MPI communicator containing the data

    Returns
    -------

    w : list of array, element i has shape=[voxels_i, features]
        The initialized orthogonal transforms (mappings) :math:`W_i` for each
        subject.

    voxels : list of int
        A list with the number of voxels per subject.


    Note
    ----

        This function assumes that the numpy random number generator was
        initialized.

        Not thread safe.
    """
    w = []
    subjects = len(data)
    voxels = np.empty(subjects, dtype=int)

    # Set Wi to a random orthogonal voxels by features matrix
    for subject in range(subjects):
        if data[subject] is not None:
            voxels[subject] = data[subject].shape[0]
            rnd_matrix = random_states[subject].random_sample((
                voxels[subject], features))
            q, r = np.linalg.qr(rnd_matrix)
            w.append(q)
        else:
            voxels[subject] = 0
            w.append(None)
    voxels = comm.allreduce(voxels, op=MPI.SUM)
    return w, voxels


def load(file):
    """Load fitted SRM from .npz file.

    Parameters
    ----------

    file : str, file-like object, or pathlib.Path
        The .npz file to read containing fitted SRM saved using srm.save

    Returns
    --------

    srm : fitted SRM model
    """

    # Load file and extract SRM attributes
    loaded = np.load(file)
    w_ = [s for s in loaded['w_']]
    s_ = loaded['s_']
    sigma_s_ = loaded['sigma_s_']
    mu_ = [s for s in loaded['mu_']]
    rho2_ = loaded['rho2_']
    features, n_iter, rand_seed = loaded['kwargs']

    # Initialize new SRM object and attach loaded attributes
    srm = SRM(n_iter=n_iter, features=features, rand_seed=rand_seed)
    srm.w_ = w_
    srm.s_ = s_
    srm.sigma_s_ = sigma_s_
    srm.mu_ = mu_
    srm.rho2_ = rho2_

    return srm


class SRM(BaseEstimator, TransformerMixin):
    """Probabilistic Shared Response Model (SRM)

    Given multi-subject data, factorize it as a shared response S among all
    subjects and an orthogonal transform W per subject:

    .. math:: X_i \\approx W_i S, \\forall i=1 \\dots N

    Parameters
    ----------

    n_iter : int, default: 10
        Number of iterations to run the algorithm.

    features : int, default: 50
        Number of features to compute.

    rand_seed : int, default: 0
        Seed for initializing the random number generator.

    comm : mpi4py.MPI.Intracomm
        The MPI communicator containing the data

    Attributes
    ----------

    w_ : list of array, element i has shape=[voxels_i, features]
        The orthogonal transforms (mappings) for each subject.

    s_ : array, shape=[features, samples]
        The shared response.

    sigma_s_ : array, shape=[features, features]
        The covariance of the shared response Normal distribution.

    mu_ : list of array, element i has shape=[voxels_i]
        The voxel means over the samples for each subject.

    rho2_ : array, shape=[subjects]
        The estimated noise variance :math:`\\rho_i^2` for each subject

    comm : mpi4py.MPI.Intracomm
        The MPI communicator containing the data

    random_state_: `RandomState`
        Random number generator initialized using rand_seed

    Note
    ----

       The number of voxels may be different between subjects. However, the
       number of samples must be the same across subjects.

       The probabilistic Shared Response Model is approximated using the
       Expectation Maximization (EM) algorithm proposed in [Chen2015]_. The
       implementation follows the optimizations published in [Anderson2016]_.

       This is a single node version.

       The run-time complexity is :math:`O(I (V T K + V K^2 + K^3))` and the
       memory complexity is :math:`O(V T)` with I - the number of iterations,
       V - the sum of voxels from all subjects, T - the number of samples, and
       K - the number of features (typically, :math:`V \\gg T \\gg K`).
    """

    def __init__(self, n_iter=10, features=50, rand_seed=0,
                 comm=MPI.COMM_SELF):
        self.n_iter = n_iter
        self.features = features
        self.rand_seed = rand_seed
        self.comm = comm
        return

    def fit(self, X, y=None):
        """Compute the probabilistic Shared Response Model

        Parameters
        ----------
        X :  list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        y : not used
        """
        logger.info('Starting Probabilistic SRM')

        # Check the number of subjects
        if len(X) <= 1:
            raise ValueError("There are not enough subjects "
                             "({0:d}) to train the model.".format(len(X)))

        # Check for input data sizes
        number_subjects = len(X)
        number_subjects_vec = self.comm.allgather(number_subjects)
        for rank in range(self.comm.Get_size()):
            if number_subjects_vec[rank] != number_subjects:
                raise ValueError(
                    "Not all ranks have same number of subjects")

        # Collect size information
        shape0 = np.zeros((number_subjects,), dtype=int)
        shape1 = np.zeros((number_subjects,), dtype=int)

        for subject in range(number_subjects):
            if X[subject] is not None:
                assert_all_finite(X[subject])
                shape0[subject] = X[subject].shape[0]
                shape1[subject] = X[subject].shape[1]

        shape0 = self.comm.allreduce(shape0, op=MPI.SUM)
        shape1 = self.comm.allreduce(shape1, op=MPI.SUM)

        # Check if all subjects have same number of TRs
        number_trs = np.min(shape1)
        for subject in range(number_subjects):
            if shape1[subject] < self.features:
                raise ValueError(
                    "There are not enough samples to train the model with "
                    "{0:d} features.".format(self.features))
            if shape1[subject] != number_trs:
                raise ValueError("Different number of samples between subjects"
                                 ".")
        # Run SRM
        self.sigma_s_, self.w_, self.mu_, self.rho2_, self.s_ = self._srm(X)

        return self

    def transform(self, X, y=None):
        """Use the model to transform matrix to Shared Response space

        Parameters
        ----------
        X : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject
            note that number of voxels and samples can vary across subjects
        y : not used (as it is unsupervised learning)


        Returns
        -------
        s : list of 2D arrays, element i has shape=[features_i, samples_i]
            Shared responses from input data (X)
        """

        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of subjects
        if len(X) != len(self.w_):
            raise ValueError("The number of subjects does not match the one"
                             " in the model.")

        s = [None] * len(X)
        for subject in range(len(X)):
            if X[subject] is not None:
                s[subject] = self.w_[subject].T.dot(X[subject])

        return s

    def _init_structures(self, data, subjects):
        """Initializes data structures for SRM and preprocess the data.


        Parameters
        ----------
        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        subjects : int
            The total number of subjects in `data`.


        Returns
        -------
        x : list of array, element i has shape=[voxels_i, samples]
            Demeaned data for each subject.

        mu : list of array, element i has shape=[voxels_i]
            Voxel means over samples, per subject.

        rho2 : array, shape=[subjects]
            Noise variance :math:`\\rho^2` per subject.

        trace_xtx : array, shape=[subjects]
            The squared Frobenius norm of the demeaned data in `x`.
        """
        x = []
        mu = []
        rho2 = np.zeros(subjects)

        trace_xtx = np.zeros(subjects)

        for subject in range(subjects):
            rho2[subject] = 1
            if data[subject] is not None:
                mu.append(np.mean(data[subject], 1))
                trace_xtx[subject] = np.sum(data[subject] ** 2)
                x.append(data[subject] - mu[subject][:, np.newaxis])
            else:
                mu.append(None)
                trace_xtx[subject] = 0
                x.append(None)

        return x, mu, rho2, trace_xtx

    def _likelihood(self, chol_sigma_s_rhos, log_det_psi, chol_sigma_s,
                    trace_xt_invsigma2_x, inv_sigma_s_rhos, wt_invpsi_x,
                    samples):
        """Calculate the log-likelihood function


        Parameters
        ----------

        chol_sigma_s_rhos : array, shape=[features, features]
            Cholesky factorization of the matrix (Sigma_S + sum_i(1/rho_i^2)
            * I)

        log_det_psi : float
            Determinant of diagonal matrix Psi (containing the rho_i^2 value
            voxels_i times).

        chol_sigma_s : array, shape=[features, features]
            Cholesky factorization of the matrix Sigma_S

        trace_xt_invsigma2_x : float
            Trace of :math:`\\sum_i (||X_i||_F^2/\\rho_i^2)`

        inv_sigma_s_rhos : array, shape=[features, features]
            Inverse of :math:`(\\Sigma_S + \\sum_i(1/\\rho_i^2) * I)`

        wt_invpsi_x : array, shape=[features, samples]

        samples : int
            The total number of samples in the data.


        Returns
        -------

        loglikehood : float
            The log-likelihood value.
        """
        log_det = (np.log(np.diag(chol_sigma_s_rhos) ** 2).sum() + log_det_psi
                   + np.log(np.diag(chol_sigma_s) ** 2).sum())
        loglikehood = -0.5 * samples * log_det - 0.5 * trace_xt_invsigma2_x
        loglikehood += 0.5 * np.trace(
            wt_invpsi_x.T.dot(inv_sigma_s_rhos).dot(wt_invpsi_x))
        # + const --> -0.5*nTR*nvoxel*subjects*math.log(2*math.pi)

        return loglikehood

    @staticmethod
    def _update_transform_subject(Xi, S):
        """Updates the mappings `W_i` for one subject.

        Parameters
        ----------

        Xi : array, shape=[voxels, timepoints]
            The fMRI data :math:`X_i` for aligning the subject.

        S : array, shape=[features, timepoints]
            The shared response.

        Returns
        -------

        Wi : array, shape=[voxels, features]
            The orthogonal transform (mapping) :math:`W_i` for the subject.
        """
        A = Xi.dot(S.T)
        # Solve the Procrustes problem
        U, _, V = np.linalg.svd(A, full_matrices=False)
        return U.dot(V)

    def transform_subject(self, X):
        """Transform a new subject using the existing model.
        The subject is assumed to have recieved equivalent stimulation

        Parameters
        ----------

        X : 2D array, shape=[voxels, timepoints]
            The fMRI data of the new subject.

        Returns
        -------

        w : 2D array, shape=[voxels, features]
            Orthogonal mapping `W_{new}` for new subject

        """
        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of TRs in the subject
        if X.shape[1] != self.s_.shape[1]:
            raise ValueError("The number of timepoints(TRs) does not match the"
                             "one in the model.")

        w = self._update_transform_subject(X, self.s_)

        return w

    def save(self, file):
        """Save fitted SRM to .npz file.

        Parameters
        ----------

        file : str, file-like object, or pathlib.Path
            Filename (string), open file (file-like object) or pathlib.Path
            where the fitted SRM will be saved. If file is a string or a Path,
            the .npz extension will be appended to the filename if it is not
            already there.

        Returns
        -------

        None
        """

        # Check if the model has been estimated
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        np.savez_compressed(
            file,
            w_=self.w_,
            s_=self.s_,
            sigma_s_=self.sigma_s_,
            mu_=self.mu_,
            rho2_=self.rho2_,
            kwargs=np.array([self.features, self.n_iter, self.rand_seed])
        )

    def _srm(self, data):
        """Expectation-Maximization algorithm for fitting the probabilistic
        SRM.

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.


        Returns
        -------

        sigma_s : array, shape=[features, features]
            The covariance :math:`\\Sigma_s` of the shared response Normal
            distribution.

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        mu : list of array, element i has shape=[voxels_i]
            The voxel means :math:`\\mu_i` over the samples for each subject.

        rho2 : array, shape=[subjects]
            The estimated noise variance :math:`\\rho_i^2` for each subject

        s : array, shape=[features, samples]
            The shared response.
        """

        local_min = min([d.shape[1] for d in data if d is not None],
                        default=sys.maxsize)
        samples = self.comm.allreduce(local_min, op=MPI.MIN)
        subjects = len(data)
        self.random_state_ = np.random.RandomState(self.rand_seed)
        random_states = [
            np.random.RandomState(self.random_state_.randint(2 ** 32,
                                                             dtype=np.int64))
            for i in range(len(data))]

        # Initialization step: initialize the outputs with initial values,
        # voxels with the number of voxels in each subject, and trace_xtx with
        # the ||X_i||_F^2 of each subject.
        w, voxels = _init_w_transforms(data, self.features, random_states,
                                       self.comm)
        x, mu, rho2, trace_xtx = self._init_structures(data, subjects)
        shared_response = np.zeros((self.features, samples))
        sigma_s = np.identity(self.features)

        rank = self.comm.Get_rank()

        # Main loop of the algorithm (run
        for iteration in range(self.n_iter):
            logger.info('Iteration %d' % (iteration + 1))

            # E-step:

            # Sum the inverted the rho2 elements for computing W^T * Psi^-1 * W
            if rank == 0:
                rho0 = (1 / rho2).sum()

                # Invert Sigma_s using Cholesky factorization
                (chol_sigma_s, lower_sigma_s) = scipy.linalg.cho_factor(
                    sigma_s, check_finite=False)
                inv_sigma_s = scipy.linalg.cho_solve(
                    (chol_sigma_s, lower_sigma_s), np.identity(self.features),
                    check_finite=False)

                # Invert (Sigma_s + rho_0 * I) using Cholesky factorization
                sigma_s_rhos = inv_sigma_s + np.identity(self.features) * rho0
                chol_sigma_s_rhos, lower_sigma_s_rhos = \
                    scipy.linalg.cho_factor(sigma_s_rhos,
                                            check_finite=False)
                inv_sigma_s_rhos = scipy.linalg.cho_solve(
                    (chol_sigma_s_rhos, lower_sigma_s_rhos),
                    np.identity(self.features), check_finite=False)

            # Compute the sum of W_i^T * rho_i^-2 * X_i, and the sum of traces
            # of X_i^T * rho_i^-2 * X_i
            wt_invpsi_x = np.zeros((self.features, samples))
            trace_xt_invsigma2_x = 0.0
            for subject in range(subjects):
                if data[subject] is not None:
                    wt_invpsi_x += (w[subject].T.dot(x[subject])) \
                                   / rho2[subject]
                    trace_xt_invsigma2_x += trace_xtx[subject] / rho2[subject]

            wt_invpsi_x = self.comm.reduce(wt_invpsi_x, op=MPI.SUM)
            trace_xt_invsigma2_x = self.comm.reduce(trace_xt_invsigma2_x,
                                                    op=MPI.SUM)
            trace_sigma_s = None
            if rank == 0:
                log_det_psi = np.sum(np.log(rho2) * voxels)

                # Update the shared response
                shared_response = sigma_s.dot(
                    np.identity(self.features) - rho0 * inv_sigma_s_rhos).dot(
                    wt_invpsi_x)

                # M-step

                # Update Sigma_s and compute its trace
                sigma_s = (inv_sigma_s_rhos
                           + shared_response.dot(shared_response.T) / samples)
                trace_sigma_s = samples * np.trace(sigma_s)

            shared_response = self.comm.bcast(shared_response)
            trace_sigma_s = self.comm.bcast(trace_sigma_s)

            # Update each subject's mapping transform W_i and error variance
            # rho_i^2
            for subject in range(subjects):
                if x[subject] is not None:
                    a_subject = x[subject].dot(shared_response.T)
                    perturbation = np.zeros(a_subject.shape)
                    np.fill_diagonal(perturbation, 0.001)
                    u_subject, s_subject, v_subject = np.linalg.svd(
                        a_subject + perturbation, full_matrices=False)
                    w[subject] = u_subject.dot(v_subject)
                    rho2[subject] = trace_xtx[subject]
                    rho2[subject] += -2 * np.sum(w[subject] * a_subject).sum()
                    rho2[subject] += trace_sigma_s
                    rho2[subject] /= samples * voxels[subject]
                else:
                    rho2[subject] = 0

            rho2 = self.comm.allreduce(rho2, op=MPI.SUM)

            if rank == 0:
                if logger.isEnabledFor(logging.INFO):
                    # Calculate and log the current log-likelihood for checking
                    # convergence
                    loglike = self._likelihood(
                        chol_sigma_s_rhos, log_det_psi, chol_sigma_s,
                        trace_xt_invsigma2_x, inv_sigma_s_rhos, wt_invpsi_x,
                        samples)
                    logger.info('Objective function %f' % loglike)

        sigma_s = self.comm.bcast(sigma_s)
        return sigma_s, w, mu, rho2, shared_response


class DetSRM(BaseEstimator, TransformerMixin):
    """Deterministic Shared Response Model (DetSRM)

    Given multi-subject data, factorize it as a shared response S among all
    subjects and an orthogonal transform W per subject:

    .. math:: X_i \\approx W_i S, \\forall i=1 \\dots N

    Parameters
    ----------

    n_iter : int, default: 10
        Number of iterations to run the algorithm.

    features : int, default: 50
        Number of features to compute.

    rand_seed : int, default: 0
        Seed for initializing the random number generator.


    Attributes
    ----------

    w_ : list of array, element i has shape=[voxels_i, features]
        The orthogonal transforms (mappings) for each subject.

    s_ : array, shape=[features, samples]
        The shared response.

    random_state_: `RandomState`
        Random number generator initialized using rand_seed

    Note
    ----

        The number of voxels may be different between subjects. However, the
        number of samples must be the same across subjects.

        The Deterministic Shared Response Model is approximated using the
        Block Coordinate Descent (BCD) algorithm proposed in [Chen2015]_.

        This is a single node version.

        The run-time complexity is :math:`O(I (V T K + V K^2))` and the memory
        complexity is :math:`O(V T)` with I - the number of iterations, V - the
        sum of voxels from all subjects, T - the number of samples, K - the
        number of features (typically, :math:`V \\gg T \\gg K`), and N - the
        number of subjects.
    """

    def __init__(self, n_iter=10, features=50, rand_seed=0):
        self.n_iter = n_iter
        self.features = features
        self.rand_seed = rand_seed
        return

    def fit(self, X, y=None):
        """Compute the Deterministic Shared Response Model

        Parameters
        ----------
        X : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        y : not used
        """
        logger.info('Starting Deterministic SRM')

        # Check the number of subjects
        if len(X) <= 1:
            raise ValueError("There are not enough subjects "
                             "({0:d}) to train the model.".format(len(X)))

        # Check for input data sizes
        if X[0].shape[1] < self.features:
            raise ValueError(
                "There are not enough samples to train the model with "
                "{0:d} features.".format(self.features))

        # Check if all subjects have same number of TRs
        number_trs = X[0].shape[1]
        number_subjects = len(X)
        for subject in range(number_subjects):
            assert_all_finite(X[subject])
            if X[subject].shape[1] != number_trs:
                raise ValueError("Different number of samples between subjects"
                                 ".")

        # Run SRM
        self.w_, self.s_ = self._srm(X)

        return self

    def transform(self, X, y=None):
        """Use the model to transform data to the Shared Response subspace

        Parameters
        ----------
        X : list of 2D arrays, element i has shape=[voxels_i, samples_i]
            Each element in the list contains the fMRI data of one subject.

        y : not used


        Returns
        -------
        s : list of 2D arrays, element i has shape=[features_i, samples_i]
            Shared responses from input data (X)
        """

        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of subjects
        if len(X) != len(self.w_):
            raise ValueError("The number of subjects does not match the one"
                             " in the model.")

        s = [None] * len(X)
        for subject in range(len(X)):
            s[subject] = self.w_[subject].T.dot(X[subject])

        return s

    def _objective_function(self, data, w, s):
        """Calculate the objective function

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        w : list of 2D arrays, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        s : array, shape=[features, samples]
            The shared response

        Returns
        -------

        objective : float
            The objective function value.
        """
        subjects = len(data)
        objective = 0.0
        for m in range(subjects):
            objective += \
                np.linalg.norm(data[m] - w[m].dot(s), 'fro') ** 2

        return objective * 0.5 / data[0].shape[1]

    def _compute_shared_response(self, data, w):
        """ Compute the shared response S

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.

        w : list of 2D arrays, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        Returns
        -------

        s : array, shape=[features, samples]
            The shared response for the subjects data with the mappings in w.
        """
        s = np.zeros((w[0].shape[1], data[0].shape[1]))
        for m in range(len(w)):
            s = s + w[m].T.dot(data[m])
        s /= len(w)

        return s

    @staticmethod
    def _update_transform_subject(Xi, S):
        """Updates the mappings `W_i` for one subject.

        Parameters
        ----------

        Xi : array, shape=[voxels, timepoints]
            The fMRI data :math:`X_i` for aligning the subject.

        S : array, shape=[features, timepoints]
            The shared response.

        Returns
        -------

        Wi : array, shape=[voxels, features]
            The orthogonal transform (mapping) :math:`W_i` for the subject.
        """
        A = Xi.dot(S.T)
        # Solve the Procrustes problem
        U, _, V = np.linalg.svd(A, full_matrices=False)
        return U.dot(V)

    def transform_subject(self, X):
        """Transform a new subject using the existing model.
        The subject is assumed to have recieved equivalent stimulation

        Parameters
        ----------

        X : 2D array, shape=[voxels, timepoints]
            The fMRI data of the new subject.

        Returns
        -------

        w : 2D array, shape=[voxels, features]
            Orthogonal mapping `W_{new}` for new subject
        """
        # Check if the model exist
        if hasattr(self, 'w_') is False:
            raise NotFittedError("The model fit has not been run yet.")

        # Check the number of TRs in the subject
        if X.shape[1] != self.s_.shape[1]:
            raise ValueError("The number of timepoints(TRs) does not match the"
                             "one in the model.")

        w = self._update_transform_subject(X, self.s_)

        return w

    def _srm(self, data):
        """Expectation-Maximization algorithm for fitting the probabilistic
        SRM.

        Parameters
        ----------

        data : list of 2D arrays, element i has shape=[voxels_i, samples]
            Each element in the list contains the fMRI data of one subject.


        Returns
        -------

        w : list of array, element i has shape=[voxels_i, features]
            The orthogonal transforms (mappings) :math:`W_i` for each subject.

        s : array, shape=[features, samples]
            The shared response.
        """

        subjects = len(data)

        self.random_state_ = np.random.RandomState(self.rand_seed)
        random_states = [
            np.random.RandomState(self.random_state_.randint(2 ** 32,
                                                             dtype=np.int64))
            for i in range(len(data))]

        # Initialization step: initialize the outputs with initial values,
        # voxels with the number of voxels in each subject.
        w, _ = _init_w_transforms(data, self.features, random_states)
        shared_response = self._compute_shared_response(data, w)
        if logger.isEnabledFor(logging.INFO):
            # Calculate the current objective function value
            objective = self._objective_function(data, w, shared_response)
            logger.info('Objective function %f' % objective)

        # Main loop of the algorithm
        for iteration in range(self.n_iter):
            logger.info('Iteration %d' % (iteration + 1))

            # Update each subject's mapping transform W_i:
            for subject in range(subjects):
                a_subject = data[subject].dot(shared_response.T)
                perturbation = np.zeros(a_subject.shape)
                np.fill_diagonal(perturbation, 0.001)
                u_subject, _, v_subject = np.linalg.svd(
                    a_subject + perturbation, full_matrices=False)
                w[subject] = u_subject.dot(v_subject)

            # Update the shared response:
            shared_response = self._compute_shared_response(data, w)

            if logger.isEnabledFor(logging.INFO):
                # Calculate the current objective function value
                objective = self._objective_function(data, w, shared_response)
                logger.info('Objective function %f' % objective)

        return w, shared_response



#  Copyright 2020 Princeton University
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Event segmentation using a Hidden Markov Model

Given an ROI timeseries, this class uses an annealed fitting procedure to
segment the timeseries into events with stable activity patterns. After
learning the signature activity pattern of each event, the model can then be
applied to other datasets to identify a corresponding sequence of events.

Full details are available in:
Christopher Baldassano, Janice Chen, Asieh Zadbood,
Jonathan W Pillow, Uri Hasson, Kenneth A Norman
Discovering event structure in continuous narrative perception and memory
Neuron, Volume 95, Issue 3, 709 - 721.e5
https://doi.org/10.1016/j.neuron.2017.06.041

This class also extends the model described in the Neuron paper:
1) It allows transition matrices that are composed of multiple separate
chains of events rather than a single linear path. This allows a model to
contain patterns for multiple event sequences (e.g. narratives), and
fit probabilities along each of these chains on a new, unlabeled timeseries.
To use this option, pass in an event_chain vector labeling which events
belong to each chain, define event patterns using set_event_patterns(),
then fit to a new dataset with find_events.

2) To obtain better fits when the underlying event structure contains
events that vary substantially in length, the split_merge option allows
the fit() function to re-distribute events during fitting. The number of
merge/split proposals is controlled by split_merge_proposals, which
controls how thorough versus fast the fitting process is.
"""

# Authors: Chris Baldassano and Cătălin Iordan (Princeton University)

import numpy as np
from scipy import stats
import logging
import copy
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted, check_array
from sklearn.exceptions import NotFittedError
import itertools

from . import _utils as utils  # type: ignore

logger = logging.getLogger(__name__)

__all__ = [
    "EventSegment",
]


class EventSegment(BaseEstimator):
    """Class for event segmentation of continuous fMRI data

    Parameters
    ----------
    n_events: int
        Number of segments to learn

    step_var: Callable[[int], float] : default 4 * (0.98 ** (step - 1))
        The Gaussian variance to use during fitting, as a function of the
        number of steps. Should decrease slowly over time.

    n_iter: int, default: 500
        Maximum number of steps to run during fitting

    event_chains: ndarray with length = n_events
        Array with unique value for each separate chain of events, each linked
        in the order they appear in the array

    split_merge: bool, default: False
        Determines whether merge/split proposals are used during fitting with
        fit(). This can improve fitting performance when events are highly
        uneven in size, but requires additional time

    split_merge_proposals: int, default: 1
        Number of merges and splits to consider at each step. Computation time
        scales as O(proposals^2) so this should usually be a small value

    Attributes
    ----------
    p_start, p_end: length n_events+1 ndarray
        initial and final prior distributions over events

    P: n_events+1 by n_events+1 ndarray
        HMM transition matrix

    ll_ : ndarray with length = number of training datasets
        Log-likelihood for training datasets over the course of training

    segments_:  list of (time by event) ndarrays
        Learned (soft) segmentation for training datasets

    event_var_ : float
        Gaussian variance at the end of learning

    event_pat_ : voxel by event ndarray
        Learned mean patterns for each event
    """

    def _default_var_schedule(step):
        return 4 * (0.98 ** (step - 1))

    def __init__(self, n_events=2,
                 step_var=_default_var_schedule,
                 n_iter=500, event_chains=None,
                 split_merge=False, split_merge_proposals=1):
        self.n_events = n_events
        self.step_var = step_var
        self.n_iter = n_iter
        self.split_merge = split_merge
        self.split_merge_proposals = split_merge_proposals
        if event_chains is None:
            self.event_chains = np.zeros(n_events)
        else:
            self.event_chains = event_chains

    def _fit_validate(self, X):
        """Validate input to fit()

        Validate data passed to fit(). Includes a transpose operation to
        change the row/column order of X and z-scoring in time.

        Parameters
        ----------
        X: time by voxel ndarray, or a list of such ndarrays
            fMRI data to be segmented

        Returns
        -------
        X: list of voxel by time ndarrays
        """
        if len(np.unique(self.event_chains)) > 1:
            raise RuntimeError("Cannot fit chains, use set_event_patterns")

        # Copy X into a list and transpose
        X = copy.deepcopy(X)
        if type(X) is not list:
            X = [X]
        for i in range(len(X)):
            X[i] = check_array(X[i])
            X[i] = X[i].T

        # Check that number of voxels is consistent across datasets
        n_dim = X[0].shape[0]
        for i in range(len(X)):
            assert (X[i].shape[0] == n_dim)

        # Double-check that data is z-scored in time
        for i in range(len(X)):
            X[i] = stats.zscore(X[i], axis=1, ddof=1)

        return X

    def fit(self, X, y=None):
        """Learn a segmentation on training data

        Fits event patterns and a segmentation to training data. After
        running this function, the learned event patterns can be used to
        segment other datasets using find_events

        Parameters
        ----------
        X: time by voxel ndarray, or a list of such ndarrays
            fMRI data to be segmented. If a list is given, then all datasets
            are segmented simultaneously with the same event patterns

        y: not used (added to comply with BaseEstimator definition)

        Returns
        -------
        self: the EventSegment object
        """

        X = self._fit_validate(X)
        n_train = len(X)
        n_dim = X[0].shape[0]
        self.classes_ = np.arange(self.n_events)

        # Initialize variables for fitting
        log_gamma = []
        for i in range(n_train):
            log_gamma.append(np.zeros((X[i].shape[1], self.n_events)))
        step = 1
        best_ll = float("-inf")
        self.ll_ = np.empty((0, n_train))
        while step <= self.n_iter:
            iteration_var = self.step_var(step)

            # Based on the current segmentation, compute the mean pattern
            # for each event
            seg_prob = [np.exp(lg) / np.sum(np.exp(lg), axis=0)
                        for lg in log_gamma]
            mean_pat = np.empty((n_train, n_dim, self.n_events))
            for i in range(n_train):
                mean_pat[i, :, :] = X[i].dot(seg_prob[i])
            mean_pat = np.mean(mean_pat, axis=0)

            # Based on the current mean patterns, compute the event
            # segmentation
            self.ll_ = np.append(self.ll_, np.empty((1, n_train)), axis=0)
            for i in range(n_train):
                logprob = self._logprob_obs(X[i], mean_pat, iteration_var)
                log_gamma[i], self.ll_[-1, i] = self._forward_backward(logprob)

            if step > 1 and self.split_merge:
                curr_ll = np.mean(self.ll_[-1, :])
                self.ll_[-1, :], log_gamma, mean_pat = \
                    self._split_merge(X, log_gamma, iteration_var, curr_ll)

            # If log-likelihood has started decreasing, undo last step and stop
            if np.mean(self.ll_[-1, :]) < best_ll:
                self.ll_ = self.ll_[:-1, :]
                break

            self.segments_ = [np.exp(lg) for lg in log_gamma]
            self.event_var_ = iteration_var
            self.event_pat_ = mean_pat
            best_ll = np.mean(self.ll_[-1, :])
            logger.debug("Fitting step %d, LL=%f", step, best_ll)

            step += 1

        return self

    def _logprob_obs(self, data, mean_pat, var):
        """Log probability of observing each timepoint under each event model

        Computes the log probability of each observed timepoint being
        generated by the Gaussian distribution for each event pattern

        Parameters
        ----------
        data: voxel by time ndarray
            fMRI data on which to compute log probabilities

        mean_pat: voxel by event ndarray
            Centers of the Gaussians for each event

        var: float or 1D array of length equal to the number of events
            Variance of the event Gaussians. If scalar, all events are
            assumed to have the same variance

        Returns
        -------
        logprob : time by event ndarray
            Log probability of each timepoint under each event Gaussian
        """

        n_vox = data.shape[0]
        t = data.shape[1]

        # z-score both data and mean patterns in space, so that Gaussians
        # are measuring Pearson correlations and are insensitive to overall
        # activity changes
        data_z = stats.zscore(data, axis=0, ddof=1)
        mean_pat_z = stats.zscore(mean_pat, axis=0, ddof=1)

        logprob = np.empty((t, self.n_events))

        if type(var) is not np.ndarray:
            var = var * np.ones(self.n_events)

        for k in range(self.n_events):
            logprob[:, k] = -0.5 * n_vox * np.log(
                2 * np.pi * var[k]) - 0.5 * np.sum(
                (data_z.T - mean_pat_z[:, k]).T ** 2, axis=0) / var[k]

        logprob /= n_vox
        return logprob

    def _forward_backward(self, logprob):
        """Runs forward-backward algorithm on observation log probs

        Given the log probability of each timepoint being generated by
        each event, run the HMM forward-backward algorithm to find the
        probability that each timepoint belongs to each event (based on the
        transition priors in p_start, p_end, and P)

        See https://en.wikipedia.org/wiki/Forward-backward_algorithm for
        mathematical details

        Parameters
        ----------
        logprob : time by event ndarray
            Log probability of each timepoint under each event Gaussian

        Returns
        -------
        log_gamma : time by event ndarray
            Log probability of each timepoint belonging to each event

        ll : float
            Log-likelihood of fit
        """
        logprob = copy.copy(logprob)
        t = logprob.shape[0]
        logprob = np.hstack((logprob, float("-inf") * np.ones((t, 1))))

        # Initialize variables
        log_scale = np.zeros(t)
        log_alpha = np.zeros((t, self.n_events + 1))
        log_beta = np.zeros((t, self.n_events + 1))

        # Set up transition matrix, with final sink state
        self.p_start = np.zeros(self.n_events + 1)
        self.p_end = np.zeros(self.n_events + 1)
        self.P = np.zeros((self.n_events + 1, self.n_events + 1))
        label_ind = np.unique(self.event_chains, return_inverse=True)[1]
        n_chains = np.max(label_ind) + 1

        # For each chain of events, link them together and then to sink state
        for c in range(n_chains):
            chain_ind = np.nonzero(label_ind == c)[0]
            self.p_start[chain_ind[0]] = 1 / n_chains
            self.p_end[chain_ind[-1]] = 1 / n_chains

            p_trans = (len(chain_ind) - 1) / t
            if p_trans >= 1:
                raise ValueError('Too few timepoints')
            for i in range(len(chain_ind)):
                self.P[chain_ind[i], chain_ind[i]] = 1 - p_trans
                if i < len(chain_ind) - 1:
                    self.P[chain_ind[i], chain_ind[i+1]] = p_trans
                else:
                    self.P[chain_ind[i], -1] = p_trans
        self.P[-1, -1] = 1

        # Forward pass
        for i in range(t):
            if i == 0:
                log_alpha[0, :] = self._log(self.p_start) + logprob[0, :]
            else:
                log_alpha[i, :] = self._log(np.exp(log_alpha[i - 1, :])
                                            .dot(self.P)) + logprob[i, :]

            log_scale[i] = np.logaddexp.reduce(log_alpha[i, :])
            log_alpha[i] -= log_scale[i]

        # Backward pass
        log_beta[-1, :] = self._log(self.p_end) - log_scale[-1]
        for i in reversed(range(t - 1)):
            obs_weighted = log_beta[i + 1, :] + logprob[i + 1, :]
            offset = np.max(obs_weighted)
            log_beta[i, :] = offset + self._log(
                np.exp(obs_weighted - offset).dot(self.P.T)) - log_scale[i]

        # Combine and normalize
        log_gamma = log_alpha + log_beta
        log_gamma -= np.logaddexp.reduce(log_gamma, axis=1, keepdims=True)

        ll = np.sum(log_scale[:(t - 1)]) + np.logaddexp.reduce(
            log_alpha[-1, :] + log_scale[-1] + self._log(self.p_end))

        log_gamma = log_gamma[:, :-1]

        return log_gamma, ll

    def _log(self, x):
        """Modified version of np.log that manually sets values <=0 to -inf

        Parameters
        ----------
        x: ndarray of floats
            Input to the log function

        Returns
        -------
        log_ma: ndarray of floats
            log of x, with x<=0 values replaced with -inf
        """

        xshape = x.shape
        _x = x.flatten()
        y = utils.masked_log(_x)
        return y.reshape(xshape)

    def set_event_patterns(self, event_pat):
        """Set HMM event patterns manually

        Rather than fitting the event patterns automatically using fit(), this
        function allows them to be set explicitly. They can then be used to
        find corresponding events in a new dataset, using find_events().

        Parameters
        ----------
        event_pat: voxel by event ndarray
        """
        if event_pat.shape[1] != self.n_events:
            raise ValueError(("Number of columns of event_pat must match "
                              "number of events"))
        self.event_pat_ = event_pat.copy()

    def find_events(self, testing_data, var=None, scramble=False):
        """Applies learned event segmentation to new testing dataset

        After fitting an event segmentation using fit() or setting event
        patterns directly using set_event_patterns(), this function finds the
        same sequence of event patterns in a new testing dataset.

        Parameters
        ----------
        testing_data: timepoint by voxel ndarray
            fMRI data to segment based on previously-learned event patterns

        var: float or 1D ndarray of length equal to the number of events
            default: uses variance that maximized training log-likelihood
            Variance of the event Gaussians. If scalar, all events are
            assumed to have the same variance. If fit() has not previously
            been run, this must be specifed (cannot be None).

        scramble: bool : default False
            If true, the order of the learned events are shuffled before
            fitting, to give a null distribution

        Returns
        -------
        segments : time by event ndarray
            The resulting soft segmentation. segments[t,e] = probability
            that timepoint t is in event e

        test_ll : float
            Log-likelihood of model fit
        """

        if var is None:
            if not hasattr(self, 'event_var_'):
                raise NotFittedError(("Event variance must be provided, if "
                                      "not previously set by fit()"))
            else:
                var = self.event_var_

        if not hasattr(self, 'event_pat_'):
            raise NotFittedError(("The event patterns must first be set "
                                  "by fit() or set_event_patterns()"))
        if scramble:
            mean_pat = self.event_pat_[:, np.random.permutation(self.n_events)]
        else:
            mean_pat = self.event_pat_

        logprob = self._logprob_obs(testing_data.T, mean_pat, var)
        lg, test_ll = self._forward_backward(logprob)
        segments = np.exp(lg)

        return segments, test_ll

    def predict(self, X):
        """Applies learned event segmentation to new testing dataset

        Alternative function for segmenting a new dataset after using
        fit() to learn a sequence of events, to comply with the sklearn
        Classifier interface

        Parameters
        ----------
        X: timepoint by voxel ndarray
            fMRI data to segment based on previously-learned event patterns

        Returns
        -------
        Event label for each timepoint
        """
        check_is_fitted(self, ["event_pat_", "event_var_"])
        X = check_array(X)
        segments, test_ll = self.find_events(X)
        return np.argmax(segments, axis=1)

    def calc_weighted_event_var(self, D, weights, event_pat):
        """Computes normalized weighted variance around event pattern

        Utility function for computing variance in a training set of weighted
        event examples. For each event, the sum of squared differences for all
        timepoints from the event pattern is computed, and then the weights
        specify how much each of these differences contributes to the
        variance (normalized by the number of voxels).

        Parameters
        ----------
        D : timepoint by voxel ndarray
            fMRI data for which to compute event variances

        weights : timepoint by event ndarray
            specifies relative weights of timepoints for each event

        event_pat : voxel by event ndarray
            mean event patterns to compute variance around

        Returns
        -------
        ev_var : ndarray of variances for each event
        """
        Dz = stats.zscore(D, axis=1, ddof=1)
        ev_var = np.empty(event_pat.shape[1])
        for e in range(event_pat.shape[1]):
            # Only compute variances for weights > 0.1% of max weight
            nz = weights[:, e] > np.max(weights[:, e])/1000
            sumsq = np.dot(weights[nz, e],
                           np.sum(np.square(Dz[nz, :] -
                                  event_pat[:, e]), axis=1))
            ev_var[e] = sumsq/(np.sum(weights[nz, e]) -
                               np.sum(np.square(weights[nz, e])) /
                               np.sum(weights[nz, e]))
        ev_var = ev_var / D.shape[1]
        return ev_var

    def model_prior(self, t):
        """Returns the prior probability of the HMM

        Runs forward-backward without any data, showing the prior distribution
        of the model (for comparison with a posterior).

        Parameters
        ----------
        t: int
            Number of timepoints

        Returns
        -------
        segments : time by event ndarray
            segments[t,e] = prior probability that timepoint t is in event e

        test_ll : float
            Log-likelihood of model (data-independent term)"""

        lg, test_ll = self._forward_backward(np.zeros((t, self.n_events)))
        segments = np.exp(lg)

        return segments, test_ll

    def _split_merge(self, X, log_gamma, iteration_var, curr_ll):
        """Attempt to improve log-likelihood with a merge/split

        The simulated annealing used in fit() is susceptible to getting
        stuck in a local minimum if there are some very short events. This
        function attempts to find
        a) pairs of neighboring events that are highly similar, to merge
        b) events that can be split into two dissimilar events
        It then tests to see whether simultaneously merging one of the
        pairs from (a) and splitting one of the events from (b) can improve
        the log-likelihood. The number of (a)/(b) pairs tested is determined
        by the split_merge_proposals class attribute.

        Parameters
        ----------
        X: list of voxel by time ndarrays
            fMRI datasets being fit

        log_gamma : list of time by event ndarrays
            Log probability of each timepoint belonging to each event,
            for each dataset

        iteration_var : float
            Current variance in simulated annealing

        curr_ll: float
            Log-likelihood of current model

        Returns
        -------
        return_ll : ndarray with length equal to length of X
            Log-likelihood after merge/split (same as curr_ll if no
            merge/split improved curr_ll)

        return_lg : list of time by event ndarrays
            Log probability of each timepoint belonging to each event,
            for each dataset (same as log_gamma if no merge/split
            improved curr_ll)

        return_mp : voxel by event ndarray
            Mean patterns of events (after possible merge/split)
        """

        # Compute current probabilities and mean patterns
        n_train = len(X)
        n_dim = X[0].shape[0]

        seg_prob = [np.exp(lg) / np.sum(np.exp(lg), axis=0)
                    for lg in log_gamma]
        mean_pat = np.empty((n_train, n_dim, self.n_events))
        for i in range(n_train):
            mean_pat[i, :, :] = X[i].dot(seg_prob[i])
        mean_pat = np.mean(mean_pat, axis=0)

        # For each event, merge its probability distribution
        # with the next event, and also split its probability
        # distribution at its median into two separate events.
        # Use these new event probability distributions to compute
        # merged and split event patterns.
        merge_pat = np.empty((n_train, n_dim, self.n_events))
        split_pat = np.empty((n_train, n_dim, 2 * self.n_events))
        for i, sp in enumerate(seg_prob):  # Iterate over datasets
            m_evprob = np.zeros((sp.shape[0], sp.shape[1]))
            s_evprob = np.zeros((sp.shape[0], 2 * sp.shape[1]))
            cs = np.cumsum(sp, axis=0)
            for e in range(sp.shape[1]):
                # Split distribution at midpoint and normalize each half
                mid = np.where(cs[:, e] >= 0.5)[0][0]
                cs_first = cs[mid, e] - sp[mid, e]
                cs_second = 1 - cs_first
                s_evprob[:mid, 2 * e] = sp[:mid, e] / cs_first
                s_evprob[mid:, 2 * e + 1] = sp[mid:, e] / cs_second

                # Merge distribution with next event distribution
                m_evprob[:, e] = sp[:, e:(e + 2)].mean(1)

            # Weight data by distribution to get event patterns
            merge_pat[i, :, :] = X[i].dot(m_evprob)
            split_pat[i, :, :] = X[i].dot(s_evprob)

        # Average across datasets
        merge_pat = np.mean(merge_pat, axis=0)
        split_pat = np.mean(split_pat, axis=0)

        # Correlate the current event patterns with the split and
        # merged patterns
        merge_corr = np.zeros(self.n_events)
        split_corr = np.zeros(self.n_events)
        for e in range(self.n_events):
            split_corr[e] = np.corrcoef(mean_pat[:, e],
                                        split_pat[:, (2 * e):(2 * e + 2)],
                                        rowvar=False)[0, 1:3].max()
            merge_corr[e] = np.corrcoef(merge_pat[:, e],
                                        mean_pat[:, e:(e + 2)],
                                        rowvar=False)[0, 1:3].min()
        merge_corr = merge_corr[:-1]

        # Find best merge/split candidates
        # A high value of merge_corr indicates that a pair of events are
        # very similar to their merged pattern, and are good candidates for
        # being merged.
        # A low value of split_corr indicates that an event's pattern is
        # very dissimilar from the patterns in its first and second half,
        # and is a good candidate for being split.
        best_merge = np.flipud(np.argsort(merge_corr))
        best_merge = best_merge[:self.split_merge_proposals]
        best_split = np.argsort(split_corr)
        best_split = best_split[:self.split_merge_proposals]

        # For every pair of merge/split candidates, attempt the merge/split
        # and measure the log-likelihood. If any are better than curr_ll,
        # accept this best merge/split
        mean_pat_last = mean_pat.copy()
        return_ll = curr_ll
        return_lg = copy.deepcopy(log_gamma)
        return_mp = mean_pat.copy()
        for m_e, s_e in itertools.product(best_merge, best_split):
            if m_e == s_e or m_e+1 == s_e:
                # Don't attempt to merge/split same event
                continue

            # Construct new set of patterns with merge/split
            mean_pat_ms = np.delete(mean_pat_last, s_e, axis=1)
            mean_pat_ms = np.insert(mean_pat_ms, [s_e, s_e],
                                    split_pat[:, (2 * s_e):(2 * s_e + 2)],
                                    axis=1)
            mean_pat_ms = np.delete(mean_pat_ms,
                                    [m_e + (s_e < m_e), m_e + (s_e < m_e) + 1],
                                    axis=1)
            mean_pat_ms = np.insert(mean_pat_ms, m_e + (s_e < m_e),
                                    merge_pat[:, m_e], axis=1)

            # Measure log-likelihood with these new patterns
            ll_ms = np.zeros(n_train)
            log_gamma_ms = list()
            for i in range(n_train):
                logprob = self._logprob_obs(X[i],
                                            mean_pat_ms, iteration_var)
                lg, ll_ms[i] = self._forward_backward(logprob)
                log_gamma_ms.append(lg)

            # If better than best ll so far, save to return to fit()
            if ll_ms.mean() > return_ll:
                return_mp = mean_pat_ms.copy()
                return_ll = ll_ms
                for i in range(n_train):
                    return_lg[i] = log_gamma_ms[i].copy()
                logger.debug("Identified merge %d,%d and split %d",
                             m_e, m_e+1, s_e)

        return return_ll, return_lg, return_mp



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Topographical Factor Analysis (TFA)

This implementation is based on the work in [Manning2014]_ and
[AndersonM2016]_.

.. [Manning2014] "Topographic factor analysis: a bayesian model for inferring
   brain networks from neural data", J. R. Manning, R. Ranganath, K. A. Norman,
   and D. M. Blei.PLoS One, vol. 9, no. 5, 2014.

.. [AndersonM2016] "Scaling Up Multi-Subject Neuroimaging Factor Analysis"
   Michael J. Anderson, Mihai Capota, Javier S. Turek, Xia Zhu,
   Theodore L. Willke, Yida Wang, Po-Hsuan Chen, Jeremy R. Manning,
   Peter J. Ramadge, and Kenneth A. Norman
   2016.
"""

# Authors: Xia Zhu (Intel Labs), Jeremy Manning (Dartmouth College) 2015~2016

from sklearn.base import BaseEstimator
from sklearn.metrics import mean_squared_error
from sklearn.cluster import KMeans
from scipy.optimize import least_squares
from scipy.optimize import linear_sum_assignment
from scipy.spatial import distance
from ..utils.utils import from_tri_2_sym, from_sym_2_tri
from . import tfa_extension  # type: ignore
import numpy as np
import math
import gc
import logging

__all__ = [
    "TFA",
]

logger = logging.getLogger(__name__)


class TFA(BaseEstimator):

    """Topographical Factor Analysis (TFA)

    Given data from one subject, factorize it as a spatial factor F and
    a weight matrix W.

    Parameters
    ----------

    max_iter : int, default: 10
        Number of iterations to run the algorithm.

    threshold : float, default: 1.0
       Tolerance for terminating the parameter estimation

    K : int, default: 50
       Number of factors to compute

    nlss_method : {'trf', 'dogbox', 'lm'}, default: 'trf'
       Non-Linear Least Square (NLSS) algorithm used by scipy.least_suqares to
       perform minimization. More information at
       http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    nlss_loss: str or callable, default: 'linear'
       Loss function used by scipy.least_squares.
       More information at
       http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    jac : {'2-point', '3-point', 'cs', callable}, default: '2-point'
       Method of computing the Jacobian matrix.
       More information at
       http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    x_scale : float or array_like or 'jac', default: 1.0
       Characteristic scale of each variable for scipy.least_suqares.
       More information at
       http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    tr_solver: {None, 'exact', 'lsmr'}, default: None
       Method for solving trust-region subproblems, relevant only for 'trf'
       and 'dogbox' methods.
       More information at
       http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.optimize.least_squares.html

    weight_method : {'rr','ols'}, default: 'rr'
       Method for estimating weight matrix W given X and F.
       'rr' means ridge regression, 'ols' means ordinary least square.

    upper_ratio : float, default: 1.8
       The upper bound of the ratio between factor's width
       and maximum sigma of scanner coordinates.

    lower_ratio : float, default: 0.02
       The lower bound of the ratio between factor's width
       and maximum sigma of scanner coordinates.

    max_num_voxel : int, default: 5000
       The maximum number of voxels to subsample.

    max_num_tr : int, default: 500
       The maximum number of TRs to subsample.

    seed : int, default: 100
       Seed for subsampling voxels and trs.

    verbose : boolean, default: False
       Verbose mode flag.


    Attributes
    ----------
    local_posterior_ : 1D array
        Local posterior on subject's centers and widths

    F_ : 2D array, in shape [n_voxel, K]
        Latent factors of the subject

    W_ : 2D array, in shape [K, n_tr]
        Weight matrix of the subject


    """

    def __init__(
            self,
            max_iter=10,
            threshold=1.0,
            K=50,
            nlss_method='trf',
            nlss_loss='soft_l1',
            jac='2-point',
            x_scale='jac',
            tr_solver=None,
            weight_method='rr',
            upper_ratio=1.8,
            lower_ratio=0.02,
            max_num_tr=500,
            max_num_voxel=5000,
            seed=100,
            verbose=False):
        self.miter = max_iter
        self.threshold = threshold
        self.K = K
        self.nlss_method = nlss_method
        self.nlss_loss = nlss_loss
        self.jac = jac
        self.x_scale = x_scale
        self.tr_solver = tr_solver
        self.weight_method = weight_method
        self.upper_ratio = upper_ratio
        self.lower_ratio = lower_ratio
        self.max_num_tr = max_num_tr
        self.max_num_voxel = max_num_voxel
        self.seed = seed
        self.verbose = verbose

    def set_K(self, K):
        """set K for the subject

        Parameters
        ----------

        K : integer
            Number of latent factor.


        Returns
        -------
        TFA
            Returns the instance itself.

        """
        self.K = K
        return self

    def set_prior(self, prior):
        """set prior for the subject

        Parameters
        ----------

        prior : 1D array, with K*(n_dim+1) elements
            Subject prior of centers and widths.


        Returns
        -------
        TFA
            Returns the instance itself.

        """
        self.local_prior = prior
        return self

    def set_seed(self, seed):
        """set seed for the subject

        Parameters
        ----------

        seed : int
            Seed for subsampling voxels and trs

        Returns
        -------
        TFA
            Returns the instance itself.

        """
        self.seed = seed
        return self

    def init_prior(self, R):
        """initialize prior for the subject

        Returns
        -------
        TFA
            Returns the instance itself.

        """
        centers, widths = self.init_centers_widths(R)
        # update prior
        prior = np.zeros(self.K * (self.n_dim + 1))
        self.set_centers(prior, centers)
        self.set_widths(prior, widths)
        self.set_prior(prior)
        return self

    def _assign_posterior(self):
        """assign posterior to prior based on Hungarian algorithm

        Returns
        -------
        TFA
            Returns the instance itself.
        """

        prior_centers = self.get_centers(self.local_prior)
        posterior_centers = self.get_centers(self.local_posterior_)
        posterior_widths = self.get_widths(self.local_posterior_)
        # linear assignment on centers
        cost = distance.cdist(prior_centers, posterior_centers, 'euclidean')
        _, col_ind = linear_sum_assignment(cost)
        # reorder centers/widths based on cost assignment
        self.set_centers(self.local_posterior_, posterior_centers[col_ind])
        self.set_widths(self.local_posterior_, posterior_widths[col_ind])
        return self

    def _converged(self):
        """Check convergence based on maximum absolute difference

        Returns
        -------

        converged : boolean
            Whether the parameter estimation converged.

        max_diff : float
            Maximum absolute difference between prior and posterior.

        """
        diff = self.local_prior - self.local_posterior_
        max_diff = np.max(np.fabs(diff))
        if self.verbose:
            _, mse = self._mse_converged()
            diff_ratio = np.sum(diff ** 2) / np.sum(self.local_posterior_ ** 2)
            logger.info(
                'tfa prior posterior max diff %f mse %f diff_ratio %f' %
                ((max_diff, mse, diff_ratio)))
        if max_diff > self.threshold:
            return False, max_diff
        else:
            return True, max_diff

    def _mse_converged(self):
        """Check convergence based on mean squared error

        Returns
        -------

        converged : boolean
            Whether the parameter estimation converged.

        mse : float
            Mean squared error between prior and posterior.

        """

        mse = mean_squared_error(self.local_prior, self.local_posterior_,
                                 multioutput='uniform_average')
        if mse > self.threshold:
            return False, mse
        else:
            return True, mse

    def get_map_offset(self):
        """Compute offset of prior/posterior


        Returns
        -------

        map_offest : 1D array
            The offset to different fields in prior/posterior

        """

        nfield = 4
        self.map_offset = np.zeros(nfield).astype(int)
        field_size = self.K * np.array([self.n_dim, 1, self.cov_vec_size, 1])
        for i in np.arange(nfield - 1) + 1:
            self.map_offset[i] = self.map_offset[i - 1] + field_size[i - 1]
        return self.map_offset

    def init_centers_widths(self, R):
        """Initialize prior of centers and widths

        Returns
        -------

        centers : 2D array, with shape [K, n_dim]
            Prior of factors' centers.

        widths : 1D array, with shape [K, 1]
            Prior of factors' widths.

        """

        kmeans = KMeans(
            init='k-means++',
            n_clusters=self.K,
            n_init=10,
            random_state=100)
        kmeans.fit(R)
        centers = kmeans.cluster_centers_
        widths = self._get_max_sigma(R) * np.ones((self.K, 1))
        return centers, widths

    def get_template(self, R):
        """Compute a template on latent factors

        Parameters
        ----------
        R : 2D array, in format [n_voxel, n_dim]
            The scanner coordinate matrix of one subject's fMRI data

        Returns
        -------
        template_prior : 1D array
            The template prior.

        template_centers_cov:  2D array,  in shape [n_dim, n_dim]
            The template on centers' covariance.

        template_widths_var: float
            The template on widths' variance
        """

        centers, widths = self.init_centers_widths(R)
        template_prior =\
            np.zeros(self.K * (self.n_dim + 2 + self.cov_vec_size))
        # template centers cov and widths var are const
        template_centers_cov = np.cov(R.T) * math.pow(self.K, -2 / 3.0)
        template_widths_var = self._get_max_sigma(R)
        centers_cov_all = np.tile(from_sym_2_tri(template_centers_cov), self.K)
        widths_var_all = np.tile(template_widths_var, self.K)
        # initial mean of centers' mean
        self.set_centers(template_prior, centers)
        self.set_widths(template_prior, widths)
        self.set_centers_mean_cov(template_prior, centers_cov_all)
        self.set_widths_mean_var(template_prior, widths_var_all)
        return template_prior, template_centers_cov, template_widths_var

    def set_centers(self, estimation, centers):
        """Set estimation on centers

        Parameters
        ----------
        estimation : 1D arrary
            Either prior or posterior estimation

        centers : 2D array, in shape [K, n_dim]
            Estimation on centers

        """
        estimation[0:self.map_offset[1]] = centers.ravel()

    def set_widths(self, estimation, widths):
        """Set estimation on widths

        Parameters
        ----------
        estimation : 1D arrary
            Either prior of posterior estimation

        widths : 2D array, in shape [K, 1]
            Estimation on widths

        """
        estimation[self.map_offset[1]:self.map_offset[2]] = widths.ravel()

    def set_centers_mean_cov(self, estimation, centers_mean_cov):
        """Set estimation on centers

        Parameters
        ----------
        estimation : 1D arrary
            Either prior of posterior estimation

        centers : 2D array, in shape [K, n_dim]
            Estimation on centers

        """
        estimation[self.map_offset[2]:self.map_offset[3]] =\
            centers_mean_cov.ravel()

    def set_widths_mean_var(self, estimation, widths_mean_var):
        """Set estimation on centers

        Parameters
        ----------
        estimation : 1D arrary
            Either prior of posterior estimation

        centers : 2D array, in shape [K, n_dim]
            Estimation on centers

        """
        estimation[self.map_offset[3]:] = widths_mean_var.ravel()

    def get_centers(self, estimation):
        """Get estimation on centers

        Parameters
        ----------
        estimation : 1D arrary
            Either prior of posterior estimation


        Returns
        -------
        centers : 2D array, in shape [K, n_dim]
            Estimation on centers

        """
        centers = estimation[0:self.map_offset[1]]\
            .reshape(self.K, self.n_dim)
        return centers

    def get_widths(self, estimation):
        """Get estimation on widths


        Parameters
        ----------
        estimation : 1D arrary
             Either prior of posterior estimation


        Returns
        -------
        fields : 2D array, in shape [K, 1]
            Estimation of widths

        """
        widths = estimation[self.map_offset[1]:self.map_offset[2]]\
            .reshape(self.K, 1)
        return widths

    def get_centers_mean_cov(self, estimation):
        """Get estimation on the covariance of centers' mean


        Parameters
        ----------
        estimation : 1D arrary
             Either prior of posterior estimation


        Returns
        -------

        centers_mean_cov : 2D array, in shape [K, cov_vec_size]
            Estimation of the covariance of centers' mean

        """
        centers_mean_cov = estimation[self.map_offset[2]:self.map_offset[3]]\
            .reshape(self.K, self.cov_vec_size)
        return centers_mean_cov

    def get_widths_mean_var(self, estimation):
        """Get estimation on the variance of widths' mean


        Parameters
        ----------
        estimation : 1D arrary
             Either prior of posterior estimation


        Returns
        -------

        widths_mean_var : 2D array, in shape [K, 1]
            Estimation on variance of widths' mean

        """
        widths_mean_var = \
            estimation[self.map_offset[3]:].reshape(self.K, 1)
        return widths_mean_var

    def get_factors(self, unique_R, inds, centers, widths):
        """Calculate factors based on centers and widths

        Parameters
        ----------

        unique_R : a list of array,
            Each element contains unique value in one dimension of
            scanner coordinate matrix R.

        inds : a list of array,
            Each element contains the indices to reconstruct one
            dimension of original cooridnate matrix from the unique
            array.

        centers : 2D array, with shape [K, n_dim]
            The centers of  factors.

        widths : 1D array, with shape [K, 1]
            The widths of factors.


        Returns
        -------

        F : 2D array, with shape [n_voxel,self.K]
            The latent factors from fMRI data.

        """

        F = np.zeros((len(inds[0]), self.K))
        tfa_extension.factor(
            F,
            centers,
            widths,
            unique_R[0],
            unique_R[1],
            unique_R[2],
            inds[0],
            inds[1],
            inds[2])

        return F

    def get_weights(self, data, F):
        """Calculate weight matrix based on fMRI data and factors

        Parameters
        ----------

        data : 2D array, with shape [n_voxel, n_tr]
            fMRI data from one subject

        F : 2D array, with shape [n_voxel,self.K]
            The latent factors from fMRI data.


        Returns
        -------

        W : 2D array, with shape [K, n_tr]
            The weight matrix from fMRI data.

        """

        beta = np.var(data)
        trans_F = F.T.copy()
        W = np.zeros((self.K, data.shape[1]))
        if self.weight_method == 'rr':
            W = np.linalg.solve(trans_F.dot(F) + beta * np.identity(self.K),
                                trans_F.dot(data))
        else:
            W = np.linalg.solve(trans_F.dot(F), trans_F.dot(data))
        return W

    def _get_max_sigma(self, R):
        """Calculate maximum sigma of scanner RAS coordinates

        Parameters
        ----------

        R : 2D array, with shape [n_voxel, n_dim]
            The coordinate matrix of fMRI data from one subject

        Returns
        -------

        max_sigma : float
            The maximum sigma of scanner coordinates.

        """

        max_sigma = 2.0 * math.pow(np.nanmax(np.std(R, axis=0)), 2)
        return max_sigma

    def get_bounds(self, R):
        """Calculate lower and upper bounds for centers and widths

        Parameters
        ----------

        R : 2D array, with shape [n_voxel, n_dim]
            The coordinate matrix of fMRI data from one subject


        Returns
        -------

        bounds : 2-tuple of array_like, default: None
            The lower and upper bounds on factor's centers and widths.

        """

        max_sigma = self._get_max_sigma(R)
        final_lower = np.zeros(self.K * (self.n_dim + 1))
        final_lower[0:self.K * self.n_dim] =\
            np.tile(np.nanmin(R, axis=0), self.K)
        final_lower[self.K * self.n_dim:] =\
            np.repeat(self.lower_ratio * max_sigma, self.K)
        final_upper = np.zeros(self.K * (self.n_dim + 1))
        final_upper[0:self.K * self.n_dim] =\
            np.tile(np.nanmax(R, axis=0), self.K)
        final_upper[self.K * self.n_dim:] =\
            np.repeat(self.upper_ratio * max_sigma, self.K)
        bounds = (final_lower, final_upper)
        return bounds

    def _residual_multivariate(
            self,
            estimate,
            unique_R,
            inds,
            X,
            W,
            template_centers,
            template_centers_mean_cov,
            template_widths,
            template_widths_mean_var_reci,
            data_sigma):
        """Residual function for estimating centers and widths

        Parameters
        ----------

        estimate : 1D array
            Initial estimation on centers

        unique_R : a list of array,
            Each element contains unique value in one dimension of
            coordinate matrix R.

        inds : a list of array,
            Each element contains the indices to reconstruct one
            dimension of original cooridnate matrix from the unique
            array.

        X : 2D array, with shape [n_voxel, n_tr]
            fMRI data from one subject.

        W : 2D array, with shape [K, n_tr]
            The weight matrix.

        template_centers: 2D array, with shape [K, n_dim]
            The template prior on centers

        template_centers_mean_cov: 2D array, with shape [K, cov_size]
            The template prior on covariance of centers' mean

        template_widths: 1D array
            The template prior on widths

        template_widths_mean_var_reci: 1D array
            The reciprocal of template prior on variance of widths' mean

        data_sigma: float
            The variance of X.


        Returns
        -------

        final_err : 1D array
            The residual function for estimating centers.

        """

        centers = self.get_centers(estimate)
        widths = self.get_widths(estimate)
        recon = X.size
        other_err = 0 if template_centers is None else (2 * self.K)
        final_err = np.zeros(recon + other_err)
        F = self.get_factors(unique_R, inds, centers, widths)
        sigma = np.zeros((1,))
        sigma[0] = data_sigma
        tfa_extension.recon(final_err[0:recon], X, F, W, sigma)

        if other_err > 0:
            # center error
            for k in np.arange(self.K):
                diff = (centers[k] - template_centers[k])
                cov = from_tri_2_sym(template_centers_mean_cov[k], self.n_dim)
                final_err[recon + k] = math.sqrt(
                    self.sample_scaling *
                    diff.dot(np.linalg.solve(cov, diff.T)))

            # width error
            base = recon + self.K
            dist = template_widths_mean_var_reci *\
                (widths - template_widths) ** 2
            final_err[base:] = np.sqrt(self.sample_scaling * dist).ravel()

        return final_err

    def _estimate_centers_widths(
            self,
            unique_R,
            inds,
            X,
            W,
            init_centers,
            init_widths,
            template_centers,
            template_widths,
            template_centers_mean_cov,
            template_widths_mean_var_reci):
        """Estimate centers and widths

        Parameters
        ----------

        unique_R : a list of array,
            Each element contains unique value in one dimension of
            coordinate matrix R.

        inds : a list of array,
            Each element contains the indices to reconstruct one
            dimension of original cooridnate matrix from the unique
            array.

        X : 2D array, with shape [n_voxel, n_tr]
            fMRI data from one subject.

        W : 2D array, with shape [K, n_tr]
            The weight matrix.

        init_centers : 2D array, with shape [K, n_dim]
            The initial values of centers.

        init_widths : 1D array
            The initial values of widths.

        template_centers: 1D array
            The template prior on centers

        template_widths: 1D array
            The template prior on widths

        template_centers_mean_cov: 2D array, with shape [K, cov_size]
            The template prior on centers' mean

        template_widths_mean_var_reci: 1D array
            The reciprocal of template prior on variance of widths' mean


        Returns
        -------

        final_estimate.x: 1D array
            The newly estimated centers and widths.

        final_estimate.cost: float
            The cost value.

        """
        # least_squares only accept x in 1D format
        init_estimate = np.hstack(
            (init_centers.ravel(), init_widths.ravel()))  # .copy()
        data_sigma = 1.0 / math.sqrt(2.0) * np.std(X)
        final_estimate = least_squares(
            self._residual_multivariate,
            init_estimate,
            args=(
                unique_R,
                inds,
                X,
                W,
                template_centers,
                template_widths,
                template_centers_mean_cov,
                template_widths_mean_var_reci,
                data_sigma),
            method=self.nlss_method,
            loss=self.nlss_loss,
            bounds=self.bounds,
            verbose=0,
            x_scale=self.x_scale,
            tr_solver=self.tr_solver)
        return final_estimate.x, final_estimate.cost

    def _fit_tfa(self, data, R, template_prior=None):
        """TFA main algorithm

        Parameters
        ----------

        data: 2D array, in shape [n_voxel, n_tr]
            The fMRI data from one subject.

        R : 2D array, in shape [n_voxel, n_dim]
            The voxel coordinate matrix of fMRI data

        template_prior : 1D array,
            The template prior on centers and widths.


        Returns
        -------
        TFA
            Returns the instance itself.

        """
        if template_prior is None:
            template_centers = None
            template_widths = None
            template_centers_mean_cov = None
            template_widths_mean_var_reci = None
        else:
            template_centers = self.get_centers(template_prior)
            template_widths = self.get_widths(template_prior)
            template_centers_mean_cov =\
                self.get_centers_mean_cov(template_prior)
            template_widths_mean_var_reci = 1.0 /\
                self.get_widths_mean_var(template_prior)
        inner_converged = False
        np.random.seed(self.seed)
        n = 0
        while n < self.miter and not inner_converged:
            self._fit_tfa_inner(
                data,
                R,
                template_centers,
                template_widths,
                template_centers_mean_cov,
                template_widths_mean_var_reci)
            self._assign_posterior()
            inner_converged, _ = self._converged()
            if not inner_converged:
                self.local_prior = self.local_posterior_
            else:
                logger.info("TFA converged at %d iteration." % (n))
            n += 1
            gc.collect()
        return self

    def get_unique_R(self, R):
        """Get unique vlaues from coordinate matrix

        Parameters
        ----------
        R :  2D array
           The coordinate matrix of a subject's fMRI data

        Return
        ------

        unique_R : a list of array,
            Each element contains unique value in one dimension of
            coordinate matrix R.

        inds : a list of array,
            Each element contains the indices to reconstruct one
            dimension of original cooridnate matrix from the unique
            array.

        """
        unique_R = []
        inds = []
        for d in np.arange(self.n_dim):
            tmp_unique, tmp_inds = np.unique(R[:, d], return_inverse=True)
            unique_R.append(tmp_unique)
            inds.append(tmp_inds)
        return unique_R, inds

    def _fit_tfa_inner(
            self,
            data,
            R,
            template_centers,
            template_widths,
            template_centers_mean_cov,
            template_widths_mean_var_reci):
        """Fit TFA model, the inner loop part

        Parameters
        ----------

        data: 2D array, in shape [n_voxel, n_tr]
            The fMRI data of a subject

        R : 2D array, in shape [n_voxel, n_dim]
            The voxel coordinate matrix of fMRI data

        template_centers: 1D array
            The template prior on centers

        template_widths: 1D array
            The template prior on widths

        template_centers_mean_cov: 2D array, with shape [K, cov_size]
            The template prior on covariance of centers' mean

        template_widths_mean_var_reci: 1D array
            The reciprocal of template prior on variance of widths' mean


        Returns
        -------
        TFA
            Returns the instance itself.

        """
        nfeature = data.shape[0]
        nsample = data.shape[1]
        feature_indices =\
            np.random.choice(nfeature, self.max_num_voxel, replace=False)
        sample_features = np.zeros(nfeature).astype(bool)
        sample_features[feature_indices] = True
        samples_indices =\
            np.random.choice(nsample, self.max_num_tr, replace=False)
        curr_data = np.zeros((self.max_num_voxel, self.max_num_tr))\
            .astype(float)
        curr_data = data[feature_indices]
        curr_data = curr_data[:, samples_indices].copy()
        curr_R = R[feature_indices].copy()
        centers = self.get_centers(self.local_prior)
        widths = self.get_widths(self.local_prior)
        unique_R, inds = self.get_unique_R(curr_R)
        F = self.get_factors(unique_R, inds, centers, widths)
        W = self.get_weights(curr_data, F)
        self.local_posterior_, self.total_cost = self._estimate_centers_widths(
            unique_R, inds, curr_data, W, centers, widths,
            template_centers, template_centers_mean_cov,
            template_widths, template_widths_mean_var_reci)

        return self

    def fit(self, X, R, template_prior=None):
        """ Topographical Factor Analysis (TFA)[Manning2014]

        Parameters
        ----------
        X : 2D array, in shape [n_voxel, n_sample]
            The fMRI data of one subject

        R : 2D array, in shape [n_voxel, n_dim]
            The voxel coordinate matrix of fMRI data

        template_prior : None or 1D array
            The template prior as an extra constraint
            None when fitting TFA alone
        """
        if self.verbose:
            logger.info('Start to fit TFA ')

        if not isinstance(X, np.ndarray):
            raise TypeError("Input data should be an array")
        if X.ndim != 2:
            raise TypeError("Input data should be 2D array")
        if not isinstance(R, np.ndarray):
            raise TypeError("Input coordinate matrix should be an array")
        if R.ndim != 2:
            raise TypeError("Input coordinate matrix should be 2D array")
        if X.shape[0] != R.shape[0]:
            raise TypeError(
                "The number of voxels should be the same in X and R!")
        if self.weight_method != 'rr' and self.weight_method != 'ols':
            raise ValueError(
                "only 'rr' and 'ols' are accepted as weight_method!")

        # main algorithm
        self.n_dim = R.shape[1]
        self.cov_vec_size = np.sum(np.arange(self.n_dim) + 1)
        self.map_offset = self.get_map_offset()
        self.bounds = self.get_bounds(R)
        n_voxel = X.shape[0]
        n_tr = X.shape[1]
        self.sample_scaling = 0.5 * float(
            self.max_num_voxel * self.max_num_tr) / float(n_voxel * n_tr)
        if template_prior is None:
            self.init_prior(R)
        else:
            self.local_prior = template_prior[0: self.map_offset[2]]
        self._fit_tfa(X, R, template_prior)
        if template_prior is None:
            centers = self.get_centers(self.local_posterior_)
            widths = self.get_widths(self.local_posterior_)
            unique_R, inds = self.get_unique_R(R)
            self.F_ = self.get_factors(unique_R, inds, centers, widths)
            self.W_ = self.get_weights(X, self.F_)
        return self



from brainiak.matnormal.utils import (pack_trainable_vars,
                                      unpack_trainable_vars,
                                      flatten_cholesky_unique,
                                      unflatten_cholesky_unique)
import tensorflow as tf
import numpy as np
import numpy.testing as npt


def test_pack_unpack(seeded_rng):

    shapes = [[2, 3], [3], [3, 4, 2], [1, 5]]
    mats = [tf.random.stateless_normal(
        shape=shape, seed=[0, 0]) for shape in shapes]
    flatmats = pack_trainable_vars(mats)
    unflatmats = unpack_trainable_vars(flatmats, mats)
    for mat_in, mat_out in zip(mats, unflatmats):
        assert tf.math.reduce_all(tf.equal(mat_in, mat_out))


def test_cholesky_uncholesky(seeded_rng):
    size = 3
    flat_chol_length = (size*(size+1))//2
    flatchol = np.random.normal(size=[flat_chol_length])
    unflatchol = unflatten_cholesky_unique(flatchol)
    npt.assert_equal(unflatchol.shape, [3, 3])
    reflatchol = flatten_cholesky_unique(unflatchol)
    npt.assert_allclose(flatchol, reflatchol)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest


def test_tri_sym_convert():
    from brainiak.utils.utils import from_tri_2_sym, from_sym_2_tri
    import numpy as np

    sym = np.random.rand(3, 3)
    tri = from_sym_2_tri(sym)
    assert tri.shape[0] == 6,\
        "from_sym_2_tri returned wrong result!"
    sym1 = from_tri_2_sym(tri, 3)
    assert sym1.shape[0] == sym1.shape[1],\
        "from_tri_2_sym returned wrong shape!"
    tri1 = from_sym_2_tri(sym1)
    assert np.array_equiv(tri, tri1),\
        "from_sym_2_tri returned wrong result!"


def test_sumexp():
    from brainiak.utils.utils import sumexp_stable
    import numpy as np

    data = np.array([[1, 1], [0, 1]])
    sums, maxs, exps = sumexp_stable(data)
    assert sums.size == data.shape[1], (
        "Invalid sum(exp(v)) computation (wrong # samples in sums)")
    assert exps.shape[0] == data.shape[0], (
        "Invalid exp(v) computation (wrong # features)")
    assert exps.shape[1] == data.shape[1], (
        "Invalid exp(v) computation (wrong # samples)")
    assert maxs.size == data.shape[1], (
        "Invalid max computation (wrong # samples in maxs)")


def test_concatenate_not_none():
    from brainiak.utils.utils import concatenate_not_none
    import numpy as np
    arrays = [None] * 5

    arrays[1] = np.array([0, 1, 2])
    arrays[3] = np.array([3, 4])

    r = concatenate_not_none(arrays, axis=0)

    assert np.all(np.arange(5) == r), (
        "Invalid concatenation of a list of arrays")


def test_cov2corr():
    from brainiak.utils.utils import cov2corr
    import numpy as np
    cov = np.array([[4, 3, 0], [3, 9, 0], [0, 0, 1]])
    corr = cov2corr(cov)
    assert np.allclose(corr,
                       np.array([[1, 0.5, 0], [0.5, 1, 0], [0, 0, 1]])), (
        "Converting from covariance matrix to correlation incorrect")


def test_ReadDesign():
    from brainiak.utils.utils import ReadDesign
    import numpy as np
    import os.path
    file_path = os.path.join(os.path.dirname(__file__), "example_design.1D")
    design = ReadDesign(fname=file_path, include_orth=False,
                        include_pols=False)
    assert design, 'Failed to read design matrix'
    assert design.reg_nuisance is None, \
        'Nuiance regressor is not None when include_orth and include_pols are'\
        ' both set to False'
    read = ReadDesign()
    assert read, 'Failed to initialize an instance of the class'
    design = ReadDesign(fname=file_path, include_orth=True, include_pols=True)
    assert np.size(design.cols_nuisance) == 10, \
        'Mistake in counting the number of nuiance regressors'
    assert np.size(design.cols_task) == 17, \
        'Mistake in counting the number of task conditions'
    assert (np.shape(design.reg_nuisance)[0]
            == np.shape(design.design_task)[0]
            ), 'The number of time points in nuiance regressor does not match'\
               ' that of task response'


def test_gen_design():
    from brainiak.utils.utils import gen_design
    import numpy as np
    import os.path
    files = {'FSL1': 'example_stimtime_1_FSL.txt',
             'FSL2': 'example_stimtime_2_FSL.txt',
             'AFNI1': 'example_stimtime_1_AFNI.txt',
             'AFNI2': 'example_stimtime_2_AFNI.txt'}
    for key in files.keys():
        files[key] = os.path.join(os.path.dirname(__file__), files[key])
    design1 = gen_design(stimtime_files=files['FSL1'], scan_duration=[48, 20],
                         TR=2, style='FSL')
    assert design1.shape == (34, 1), 'Returned design matrix has wrong shape'
    assert design1[24] == 0, (
        "gen_design should generated design matrix for each run separately "
        "and concatenate them.")
    design2 = gen_design(stimtime_files=[files['FSL1'], files['FSL2']],
                         scan_duration=[48, 20], TR=2, style='FSL')
    assert design2.shape == (34, 2), 'Returned design matrix has wrong shape'
    design3 = gen_design(stimtime_files=files['FSL1'], scan_duration=68, TR=2,
                         style='FSL')
    assert design3[24] != 0, (
        'design matrix should be non-zero 8 seconds after an event onset.')
    design4 = gen_design(stimtime_files=[files['FSL2']],
                         scan_duration=[48, 20], TR=2, style='FSL')
    assert np.all(np.isclose(design1 * 0.5, design4)), (
        'gen_design does not treat missing values correctly')
    design5 = gen_design(stimtime_files=[files['FSL2']],
                         scan_duration=[48, 20], TR=1)
    assert (np.abs(design4 - design5[::2])).mean() < 0.1, (
        'design matrices sampled at different frequency do not match'
        ' at corresponding time points')
    design6 = gen_design(stimtime_files=[files['AFNI1']],
                         scan_duration=[48, 20], TR=2, style='AFNI')
    assert np.all(np.isclose(design1, design6)), (
        'design matrices generated from AFNI style and FSL style do not match')
    design7 = gen_design(stimtime_files=[files['AFNI2']],
                         scan_duration=[48], TR=2, style='AFNI')
    assert np.all(design7 == 0.0), (
        'A negative stimulus onset of AFNI style should result in an all-zero'
        + ' design matrix')


def test_center_mass_exp():
    from brainiak.utils.utils import center_mass_exp
    import numpy as np

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp([1, 2])
    assert ('interval must be a tuple'
            in str(excinfo.value))

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp((1, 2, 3))
    assert ('interval must be length two'
            in str(excinfo.value))

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp((-2, -1))
    assert ('interval_left must be non-negative'
            in str(excinfo.value))

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp((-2, 3))
    assert ('interval_left must be non-negative'
            in str(excinfo.value))

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp((3, 3))
    assert ('interval_right must be bigger than interval_left'
            in str(excinfo.value))

    with pytest.raises(AssertionError) as excinfo:
        result = center_mass_exp((1, 2), -1)
    assert ('scale must be positive'
            in str(excinfo.value))

    result = center_mass_exp((0, np.inf), 2.0)
    assert np.isclose(result, 2.0), 'center of mass '\
        'incorrect for the whole distribution'
    result = center_mass_exp((1.0, 1.0+2e-10))
    assert np.isclose(result, 1.0+1e-10), 'for a small '\
        'enough interval, the center of mass should be '\
        'close to its mid-point'


def test_p_from_null():
    import numpy as np
    from brainiak.utils.utils import p_from_null

    # Create random null and observed value in tail
    null = np.random.randn(10000)
    observed = np.ceil(np.percentile(null, 97.5) * 1000) / 1000

    # Check that we catch improper side
    with pytest.raises(ValueError):
        _ = p_from_null(observed, null, side='wrong')

    # Check two-tailed p-value for observed
    p_ts = p_from_null(observed, null)
    assert np.isclose(p_ts, 0.05, atol=1e-02)

    # Check two-tailed p-value for observed
    p_right = p_from_null(observed, null, side='right')
    assert np.isclose(p_right, 0.025, atol=1e-02)
    assert np.isclose(p_right, p_ts / 2, atol=1e-02)

    # Check two-tailed p-value for observed
    p_left = p_from_null(observed, null, side='left')
    assert np.isclose(p_left, 0.975, atol=1e-02)
    assert np.isclose(1 - p_left, p_right, atol=1e-02)
    assert np.isclose(1 - p_left, p_ts / 2, atol=1e-02)

    # Check 2-dimensional input (i.e., samples by voxels)
    null = np.random.randn(10000, 3)
    observed = np.ceil(np.percentile(null, 97.5, axis=0) * 1000) / 1000

    # Check two-tailed p-value for observed
    p_ts = p_from_null(observed, null, axis=0)
    assert np.allclose(p_ts, 0.05, atol=1e-02)

    # Check two-tailed p-value for observed
    p_right = p_from_null(observed, null, side='right', axis=0)
    assert np.allclose(p_right, 0.025, atol=1e-02)
    assert np.allclose(p_right, p_ts / 2, atol=1e-02)

    # Check two-tailed p-value for observed
    p_left = p_from_null(observed, null, side='left', axis=0)
    assert np.allclose(p_left, 0.975, atol=1e-02)
    assert np.allclose(1 - p_left, p_right, atol=1e-02)
    assert np.allclose(1 - p_left, p_ts / 2, atol=1e-02)

    # Check for exact test
    p_ts = p_from_null(observed, null, exact=True, axis=0)
    assert np.allclose(p_ts, 0.05, atol=1e-02)

    # Check two-tailed p-value for exact
    p_right = p_from_null(observed, null, side='right',
                          exact=True, axis=0)
    assert np.allclose(p_right, 0.025, atol=1e-02)
    assert np.allclose(p_right, p_ts / 2, atol=1e-02)

    # Check two-tailed p-value for exact
    p_left = p_from_null(observed, null, side='left',
                         exact=True, axis=0)
    assert np.allclose(p_left, 0.975, atol=1e-02)
    assert np.allclose(1 - p_left, p_right, atol=1e-02)
    assert np.allclose(1 - p_left, p_ts / 2, atol=1e-02)


def test_phase_randomize():
    import numpy as np
    from scipy.fftpack import fft
    from scipy.stats import pearsonr
    from brainiak.utils.utils import phase_randomize

    data = np.repeat(np.repeat(np.random.randn(60)[:, np.newaxis, np.newaxis],
                               30, axis=1),
                     20, axis=2)
    assert np.array_equal(data[..., 0], data[..., 1])

    # Phase-randomize data across subjects (same across voxels)
    shifted_data = phase_randomize(data, voxelwise=False, random_state=1)
    assert shifted_data.shape == data.shape
    assert not np.array_equal(shifted_data[..., 0], shifted_data[..., 1])
    assert not np.array_equal(shifted_data[..., 0], data[..., 0])

    # Check that uneven n_TRs doesn't explode
    _ = phase_randomize(data[:-1, ...])

    # Check that random_state returns same shifts
    shifted_data_ = phase_randomize(data, voxelwise=False, random_state=1)
    assert np.array_equal(shifted_data, shifted_data_)

    shifted_data_ = phase_randomize(data, voxelwise=False, random_state=2)
    assert not np.array_equal(shifted_data, shifted_data_)

    # Phase-randomize subjects and voxels
    shifted_data = phase_randomize(data, voxelwise=True, random_state=1)
    assert shifted_data.shape == data.shape
    assert not np.array_equal(shifted_data[..., 0], shifted_data[..., 1])
    assert not np.array_equal(shifted_data[..., 0], data[..., 0])
    assert not np.array_equal(shifted_data[:, 0, 0], shifted_data[:, 1, 0])

    # Try with 2-dimensional input
    shifted_data = phase_randomize(data[..., 0],
                                   voxelwise=True,
                                   random_state=1)
    assert shifted_data.ndim == 2
    assert not np.array_equal(shifted_data[:, 0], shifted_data[:, 1])

    # Create correlated noisy data
    corr_data = np.repeat(np.random.randn(60)[:, np.newaxis, np.newaxis],
                          2, axis=2) + np.random.randn(60, 1, 2)

    # Get correlation and frequency domain for data
    corr_r = pearsonr(corr_data[:, 0, 0],
                      corr_data[:, 0, 1])[0]
    corr_freq = fft(corr_data, axis=0)

    # Phase-randomize time series and get correlation/frequency
    shifted_data = phase_randomize(corr_data)
    shifted_r = pearsonr(shifted_data[:, 0, 0],
                         shifted_data[:, 0, 1])[0]
    shifted_freq = fft(shifted_data, axis=0)

    # Check that phase-randomization reduces correlation
    assert np.abs(shifted_r) < np.abs(corr_r)

    # Check that amplitude spectrum is preserved
    assert np.allclose(np.abs(shifted_freq), np.abs(corr_freq))


def test_check_timeseries_input():
    import numpy as np
    from itertools import combinations
    from brainiak.utils.utils import _check_timeseries_input

    # Set a fixed vector for comparison
    vector = np.random.randn(60)

    # List of subjects with one voxel/ROI
    list_1d = [vector for _ in np.arange(10)]
    (data_list_1d, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(list_1d)
    assert n_TRs == 60
    assert n_voxels == 1
    assert n_subjects == 10

    # Array of subjects with one voxel/ROI
    array_2d = np.hstack([vector[:, np.newaxis]
                          for _ in np.arange(10)])
    (data_array_2d, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(array_2d)
    assert n_TRs == 60
    assert n_voxels == 1
    assert n_subjects == 10

    # List of 2-dimensional arrays
    list_2d = [vector[:, np.newaxis] for _ in np.arange(10)]
    (data_list_2d, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(list_2d)
    assert n_TRs == 60
    assert n_voxels == 1
    assert n_subjects == 10

    # Check if lists have mismatching size
    list_bad = [list_2d[0][:-1, :]] + list_2d[1:]
    with pytest.raises(ValueError):
        (data_list_bad, _, _, _) = _check_timeseries_input(list_bad)

    # List of 3-dimensional arrays
    list_3d = [vector[:, np.newaxis, np.newaxis]
               for _ in np.arange(10)]
    (data_list_3d, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(list_3d)
    assert n_TRs == 60
    assert n_voxels == 1
    assert n_subjects == 10

    # 3-dimensional array
    array_3d = np.dstack([vector[:, np.newaxis]
                          for _ in np.arange(10)])
    (data_array_3d, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(array_3d)
    assert n_TRs == 60
    assert n_voxels == 1
    assert n_subjects == 10

    # Check that 4-dimensional input array throws error
    array_4d = array_3d[..., np.newaxis]
    with pytest.raises(ValueError):
        (data_array_4d, _, _, _) = _check_timeseries_input(array_4d)

    # Check they're the same
    for pair in combinations([data_list_1d, data_array_2d,
                              data_list_2d, data_list_3d,
                              data_array_3d], 2):
        assert np.array_equal(pair[0], pair[1])

    # List of multivoxel arrays
    matrix = np.random.randn(60, 30)
    list_mv = [matrix
               for _ in np.arange(10)]
    (data_list_mv, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(list_mv)
    assert n_TRs == 60
    assert n_voxels == 30
    assert n_subjects == 10

    # 3-dimensional array with multiple voxels
    array_mv = np.dstack([matrix for _ in np.arange(10)])
    (data_array_mv, n_TRs,
     n_voxels, n_subjects) = _check_timeseries_input(array_mv)
    assert n_TRs == 60
    assert n_voxels == 30
    assert n_subjects == 10

    assert np.array_equal(data_list_mv, data_array_mv)


def test_array_correlation():
    import numpy as np
    from brainiak.utils.utils import array_correlation
    from scipy.stats import pearsonr

    # Minimal array datasets
    n_TRs = 30
    n_voxels = 2
    x, y = (np.random.randn(n_TRs, n_voxels),
            np.random.randn(n_TRs, n_voxels))

    # Perform the correlation
    r = array_correlation(x, y)

    # Check there are the right number of voxels in the output
    assert r.shape == (n_voxels,)

    # Check that this (roughly) matches corrcoef
    assert np.allclose(r, np.corrcoef(x.T, y.T)[[0, 1], [2, 3]])

    # Check that this (roughly) matches pearsonr
    assert np.allclose(r, np.array([pearsonr(x[:, 0], y[:, 0])[0],
                                    pearsonr(x[:, 1], y[:, 1])[0]]))

    # Try axis argument
    assert np.allclose(array_correlation(x, y, axis=0),
                       array_correlation(x.T, y.T, axis=1))

    # Trigger shape mismatch error
    with pytest.raises(ValueError):
        array_correlation(x, y[:, 0])

    with pytest.raises(ValueError):
        array_correlation(x, y[:-1])

    # Feed in lists
    _ = array_correlation(x.tolist(), y)
    _ = array_correlation(x, y.tolist())
    _ = array_correlation(x.tolist(), y.tolist())

    # Check 1D array input
    x, y = (np.random.randn(n_TRs),
            np.random.randn(n_TRs))

    assert type(array_correlation(x, y)) == np.float64
    assert np.isclose(array_correlation(x, y),
                      pearsonr(x, y)[0])

    # 1D list inputs
    _ = array_correlation(x.tolist(), y)
    _ = array_correlation(x, y.tolist())
    _ = array_correlation(x.tolist(), y.tolist())

    # Check integer inputs
    x, y = (np.random.randint(0, 9, (n_TRs, n_voxels)),
            np.random.randint(0, 9, (n_TRs, n_voxels)))
    _ = array_correlation(x, y)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Full correlation matrix analysis

The implementation is based on the work in [Wang2015-1]_ and [Wang2015-2]_.

.. [Wang2015-1] Full correlation matrix analysis (FCMA): An unbiased method for
   task-related functional connectivity",
   Yida Wang, Jonathan D Cohen, Kai Li, Nicholas B Turk-Browne.
   Journal of Neuroscience Methods, 2015.

.. [Wang2015-2] "Full correlation matrix analysis of fMRI data on Intel® Xeon
   Phi™ coprocessors",
   Yida Wang, Michael J. Anderson, Jonathan D. Cohen, Alexander Heinecke,
   Kai Li, Nadathur Satish, Narayanan Sundaram, Nicholas B. Turk-Browne,
   Theodore L. Willke.
   In Proceedings of the International Conference for
   High Performance Computing,
   Networking, Storage and Analysis. 2015.
"""



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import scipy.io
from scipy.stats import stats
from sklearn.metrics import confusion_matrix
from sklearn.svm import NuSVC
import numpy as np
import brainiak.funcalign.srm

# Load the input data that contains the movie stimuli for unsupervised training with SRM
movie_file = scipy.io.loadmat('data/movie_data.mat')
movie_data_left = movie_file['movie_data_lh']
movie_data_right = movie_file['movie_data_rh']
subjects = movie_data_left.shape[2]

# Convert data to a list of arrays matching SRM input.
# Each element is a matrix of voxels by TRs.
# Also, concatenate data from both hemispheres in the brain.
movie_data = []
for s in range(subjects):
    movie_data.append(np.concatenate([movie_data_left[:, :, s], movie_data_right[:, :, s]], axis=0))

# Z-score the data
for subject in range(subjects):
    movie_data[subject] = stats.zscore(movie_data[subject], axis=1, ddof=1)

# Run SRM with the movie data
help(brainiak.funcalign.srm.SRM)
srm = brainiak.funcalign.srm.SRM(n_iter=10, features=50)
srm.fit(movie_data)

# We define a function to present the output of the experiment.
def plot_confusion_matrix(cm, title="Confusion Matrix"):
    """Plots a confusion matrix for each subject
    """
    import matplotlib.pyplot as plt
    import math
    plt.figure()
    subjects = len(cm)
    root_subjects = math.sqrt(subjects)
    cols = math.ceil(root_subjects)
    rows = math.ceil(subjects/cols)
    classes = cm[0].shape[0]
    for subject in range(subjects):
        plt.subplot(rows, cols, subject+1)
        plt.imshow(cm[subject], interpolation='nearest', cmap=plt.cm.bone)
        plt.xticks(np.arange(classes), range(1, classes+1))
        plt.yticks(np.arange(classes), range(1, classes+1))
        cbar = plt.colorbar(ticks=[0.0, 1.0], shrink=0.6)
        cbar.set_clim(0.0, 1.0)
        plt.xlabel("Predicted")
        plt.ylabel("True label")
        plt.title("{0:d}".format(subject + 1))
    plt.suptitle(title)
    plt.tight_layout()
    plt.show()

# Load the input data that contains the image stimuli and its labels for training a classifier
image_file = scipy.io.loadmat('data/image_data.mat')
image_data_left = image_file['image_data_lh']
image_data_right = image_file['image_data_rh']

# Convert data to a list of arrays matching SRM input.
# Each element is a matrix of voxels by TRs.
# Also, concatenate data from both hemispheres in the brain.
image_data = []
for s in range(subjects):
    image_data.append(np.concatenate([image_data_left[:, :, s], image_data_right[:, :, s]], axis=0))

assert image_data[0].shape[0] == movie_data[0].shape[0], "Number of voxels in movie data and image data do not match!"

# Z-score the image data
for subject in range(subjects):
    image_data[subject] = stats.zscore(image_data[subject], axis=1, ddof=1)

# Z-score the shared response data
image_data_shared = srm.transform(image_data)
for subject in range(subjects):
    image_data_shared[subject] = stats.zscore(image_data_shared[subject], axis=1, ddof=1)

# Read the labels of the image data for training the classifier.
labels = scipy.io.loadmat('data/label.mat')
labels = np.squeeze(labels['label'])

# Run a leave-one-out cross validation with the subjects
train_labels = np.tile(labels, subjects-1)
test_labels = labels
accuracy = np.zeros((subjects,))
cm = [None] * subjects
for subject in range(subjects):
    # Concatenate the subjects' data for training into one matrix
    train_subjects = list(range(subjects))
    train_subjects.remove(subject)
    TRs = image_data_shared[0].shape[1]
    train_data = np.zeros((image_data_shared[0].shape[0], len(train_labels)))
    for train_subject in range(len(train_subjects)):
        start_index = train_subject*TRs
        end_index = start_index+TRs
        train_data[:, start_index:end_index] = image_data_shared[train_subjects[train_subject]]

    # Train a Nu-SVM classifier using scikit learn
    classifier = NuSVC(nu=0.5, kernel='linear', gamma='auto')
    classifier = classifier.fit(train_data.T, train_labels)

    # Predict on the test data
    predicted_labels = classifier.predict(image_data_shared[subject].T)
    accuracy[subject] = sum(predicted_labels == test_labels)/float(len(predicted_labels))

    # Create a confusion matrix to see the accuracy of each class
    cm[subject] = confusion_matrix(test_labels, predicted_labels)

    # Normalize the confusion matrix
    cm[subject] = cm[subject].astype('float') / cm[subject].sum(axis=1)[:, np.newaxis]


# Plot and print the results
plot_confusion_matrix(cm, title="Confusion matrices for different test subjects with Probabilistic SRM")
print("SRM: The average accuracy among all subjects is {0:f} +/- {1:f}".format(np.mean(accuracy), np.std(accuracy)))

# Now we repeat the experiment with the Deterministic SRM
# Fit the model
srm = brainiak.funcalign.srm.DetSRM(n_iter=10, features=50)
srm.fit(movie_data)

# Transform the image stimuli data
image_data_shared = srm.transform(image_data)
for subject in range(subjects):
    image_data_shared[subject] = stats.zscore(image_data_shared[subject], axis=1, ddof=1)

# Run a leave-one-out cross validation with the subjects
accuracy = np.zeros((subjects,))
cm = [None] * subjects
for subject in range(subjects):
    # Concatenate the subjects' data for training into one matrix
    train_subjects = list(range(subjects))
    train_subjects.remove(subject)
    TRs = image_data_shared[0].shape[1]
    train_data = np.zeros((image_data_shared[0].shape[0], len(train_labels)))
    for train_subject in range(len(train_subjects)):
        start_index = train_subject*TRs
        end_index = start_index+TRs
        train_data[:, start_index:end_index] = image_data_shared[train_subjects[train_subject]]

    # Train a Nu-SVM classifier using scikit learn
    classifier = NuSVC(nu=0.5, kernel='linear', gamma='auto')
    classifier = classifier.fit(train_data.T, train_labels)

    # Predict on the test data
    predicted_labels = classifier.predict(image_data_shared[subject].T)
    accuracy[subject] = sum(predicted_labels == test_labels)/float(len(predicted_labels))

    # Create a confusion matrix to see the accuracy of each class
    cm[subject] = confusion_matrix(test_labels, predicted_labels)

    # Normalize the confusion matrix
    cm[subject] = cm[subject].astype('float') / cm[subject].sum(axis=1)[:, np.newaxis]

# Plot and print the results
plot_confusion_matrix(cm, title="Confusion matrices for different test subjects with Deterministic SRM")
print("Det. SRM: The average accuracy among all subjects is {0:f} +/- {1:f}".format(np.mean(accuracy), np.std(accuracy)))



"""
Some properties of the matrix-variate normal distribution
---------------------------------------------------------

.. math::
    \\DeclareMathOperator{\\Tr}{Tr}
    \\newcommand{\\trp}{^{T}} % transpose
    \\newcommand{\\trace}{\\text{Trace}} % trace
    \\newcommand{\\inv}{^{-1}}
    \\newcommand{\\mb}{\\mathbf{b}}
    \\newcommand{\\M}{\\mathbf{M}}
    \\newcommand{\\C}{\\mathbf{C}}
    \\newcommand{\\G}{\\mathbf{G}}
    \\newcommand{\\A}{\\mathbf{A}}
    \\newcommand{\\R}{\\mathbf{R}}
    \\renewcommand{\\S}{\\mathbf{S}}
    \\newcommand{\\B}{\\mathbf{B}}
    \\newcommand{\\Q}{\\mathbf{Q}}
    \\newcommand{\\mH}{\\mathbf{H}}
    \\newcommand{\\U}{\\mathbf{U}}
    \\newcommand{\\mL}{\\mathbf{L}}
    \\newcommand{\\diag}{\\mathrm{diag}}
    \\newcommand{\\etr}{\\mathrm{etr}}
    \\renewcommand{\\H}{\\mathbf{H}}
    \\newcommand{\\vecop}{\\mathrm{vec}}
    \\newcommand{\\I}{\\mathbf{I}}
    \\newcommand{\\X}{\\mathbf{X}}
    \\newcommand{\\Y}{\\mathbf{Y}}
    \\newcommand{\\Z}{\\mathbf{Z}}
    \\renewcommand{\\L}{\\mathbf{L}}


The matrix-variate normal distribution is a generalization to matrices of the
normal distribution. Another name for it is the multivariate normal
distribution with kronecker separable covariance.
The distributional intuition is as follows: if
:math:`X \\sim \\mathcal{MN}(M,R,C)` then
:math:`\\mathrm{vec}(X)\\sim\\mathcal{N}(\\mathrm{vec}(M), C \\otimes R)`,
where :math:`\\mathrm{vec}(\\cdot)` is the vectorization operator and
:math:`\\otimes` is the Kronecker product. If we think of X as a matrix of TRs
by voxels in the fMRI setting, then this model assumes that each voxel has the
same TR-by-TR covariance structure (represented by the matrix R),
and each volume has the same spatial covariance (represented by the matrix C).
This assumption allows us to model both covariances separately.
We can assume that the spatial covariance itself is kronecker-structured,
which implies that the spatial covariance of voxels is the same in the X,
Y and Z dimensions.

The log-likelihood for the matrix-normal density is:

.. math::
    \\log p(X\\mid \\M,\\R, \\C) = -2\\log mn - m \\log|\\C| -  n \\log|\\R| -
    \\Tr\\left[\\C\\inv(\\X-\\M)\\trp\\R\\inv(\\X-\\M)\\right]

Here :math:`X` and :math:`M` are both :math:`m\\times n` matrices, :math:`\\R`
is :math:`m\\times m` and :math:`\\C` is :math:`n\\times n`.

The `brainiak.matnormal` package provides structure to infer models that
can be stated in the matrix-normal notation that are useful for fMRI analysis.
It provides a few interfaces. `MatnormModelBase` is intended as a
base class for matrix-variate models. It provides a wrapper for the tensorflow
optimizer that provides convergence checks based on thresholds on the function
value and gradient, and simple verbose outputs. It also provides an interface
for noise covariances (`CovBase`). Any class that follows the interface
can be used as a noise covariance in any of the matrix normal models. The
package includes a variety of noise covariances to work with.

Matrix normal marginals
-------------------------

Here we extend the multivariate gaussian marginalization identity to matrix
normals. This is used in a number of the models in the package. Below, we
use lowercase subscripts for sizes to make dimensionalities easier to track.
Uppercase subscripts for covariances help keep track where they come from.

.. math::
    \\mathbf{X}_{ij} &\\sim \\mathcal{MN}(\\mathbf{A}_{ij},
    \\Sigma_{\\mathbf{X}i},\\Sigma_{\\mathbf{X}j})\\\\
    \\mathbf{Y}_{jk} &\\sim \\mathcal{MN}(\\mathbf{B}_{jk},
     \\Sigma_{\\mathbf{Y}j},\\Sigma_{\\mathbf{Y}k})\\\\
    \\mathbf{Z}_{ik}\\mid\\mathbf{X}_{ij},\\mathbf{Y}_{jk} &\\sim
     \\mathcal{MN}(\\mathbf{X}_{ij}\\mathbf{Y}_{jk} + \\mathbf{C}_{ik},
      \\Sigma_{\\mathbf{Z}_i}, \\Sigma_{\\mathbf{Z}_k})\\\\


We vectorize, and convert to a form we recognize as
:math:`y \\sim \\mathcal{N}(Mx+b, \\Sigma)`.

.. math::
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij},\\mathbf{Y}_{jk} &\\sim
     \\mathcal{N}(\\vecop(\\X_{ij}\\mathbf{Y}_{jk}+\\mathbf{C}_{ik}),
     \\Sigma_{\\mathbf{Z}_k}\\otimes\\Sigma_{\\mathbf{Z}_i})\\\\
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij},\\mathbf{Y}_{jk}
    &\\sim \\mathcal{N}((\\I_k\\otimes\\X_{ij})\\vecop(\\mathbf{Y}_{jk})
     + \\vecop(\\mathbf{C}_{ik}),
     \\Sigma_{\\mathbf{Z}_k}\\otimes\\Sigma_{\\mathbf{Z}_i})


Now we can use our standard gaussian marginalization identity:

.. math::
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij} \\sim
    \\mathcal{N}((\\I_k\\otimes\\X_{ij})\\vecop(\\mathbf{B}_{jk}) +
     \\vecop(\\mathbf{C}_{ik}),
     \\Sigma_{\\mathbf{Z}_k}\\otimes\\Sigma_{\\mathbf{Z}_i} +
     (\\I_k\\otimes\\X_{ij})(\\Sigma_{\\mathbf{Y}_k}\\otimes
     \\Sigma_{\\mathbf{Y}_j})(\\I_k\\otimes\\X_{ij})\\trp )


Collect terms using the mixed-product property of kronecker products:

.. math::
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij} \\sim
     \\mathcal{N}(\\vecop(\\X_{ij}\\mathbf{B}_{jk}) +
      \\vecop(\\mathbf{C}_{ik}), \\Sigma_{\\mathbf{Z}_k}\\otimes
      \\Sigma_{\\mathbf{Z}_i} + \\Sigma_{\\mathbf{Y}_k}\\otimes
       \\X_{ij}\\Sigma_{\\mathbf{Y}_j}\\X_{ij}\\trp)


Now, we can see that the marginal density is a matrix-variate normal only if
:math:`\\Sigma_{\\mathbf{Z}_k}= \\Sigma_{\\mathbf{Y}_k}` -- that is, the
variable we're marginalizing over has the same covariance in the dimension
we're *not* marginalizing over as the marginal density. Otherwise the densit
is well-defined but the covariance retains its kronecker structure. So we let
:math:`\\Sigma_k:=\\Sigma_{\\mathbf{Z}_k}= \\Sigma_{\\mathbf{Y}_k}`, factor,
and transform it back into a matrix normal:

.. math::
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij} &\\sim
     \\mathcal{N}(\\vecop(\\X\\mathbf{B}_{jk}) + \\vecop(\\mathbf{C}_{ik}),
      \\Sigma_{k}\\otimes\\Sigma_{\\mathbf{Z}_i} + \\Sigma_{_k}\\otimes
      \\X\\Sigma_{\\mathbf{Y}_j}\\X\\trp)\\\\
    \\vecop(\\mathbf{Z}_{ik})\\mid\\mathbf{X}_{ij} &\\sim
     \\mathcal{N}(\\vecop(\\X\\mathbf{B}_{jk}) + \\vecop(\\mathbf{C}_{ik}),
      \\Sigma_{k}\\otimes(\\Sigma_{\\mathbf{Z}_i}
      +\\X\\Sigma_{\\mathbf{Y}_j}\\X\\trp))\\\\
    \\mathbf{Z}_{ik}\\mid\\mathbf{X}_{ij} &\\sim
     \\mathcal{MN}(\\X\\mathbf{B}_{jk} + \\mathbf{C}_{ik},
      \\Sigma_{\\mathbf{Z}_i} +\\X\\Sigma_{\\mathbf{Y}_j}\\X\\trp,\\Sigma_{k})


We can do it in the other direction as well, because if
:math:`\\X \\sim \\mathcal{MN}(M, U, V)` then :math:`\\X\\trp \\sim
\\mathcal{MN}(M\\trp, V, U)`:

.. math::
    \\mathbf{Z\\trp}_{ik}\\mid\\mathbf{X}_{ij},\\mathbf{Y}_{jk} &\\sim
    \\mathcal{MN}(\\mathbf{Y}_{jk}\\trp\\mathbf{X}_{ij}\\trp +
    \\mathbf{C}\\trp_{ik}, \\Sigma_{\\mathbf{Z}_k},\\Sigma_{\\mathbf{Z}_i})\\\\
    \\mbox{let } \\Sigma_i :=
     \\Sigma_{\\mathbf{Z}_i}=\\Sigma_{\\mathbf{X}_i} \\\\
    \\cdots\\\\
    \\mathbf{Z\\trp}_{ik}\\mid\\mathbf{Y}_{jk} &\\sim
     \\mathcal{MN}(\\mathbf{A}_{jk}\\trp\\mathbf{X}_{ij}\\trp +
      \\mathbf{C}\\trp_{ik}, \\Sigma_{\\mathbf{Z}_k} +
       \\Y\\trp\\Sigma_{\\mathbf{Y}_j}\\Y,\\Sigma_{\\mathbf{Z}_i})\\\\
    \\mathbf{Z}_{ik}\\mid\\mathbf{Y}_{jk} &\\sim
     \\mathcal{MN}(\\mathbf{X}_{ij}\\mathbf{A}_{jk}+
      \\mathbf{C}_{ik},\\Sigma_{\\mathbf{Z}_i},\\Sigma_{\\mathbf{Z}_k} +
       \\Y\\trp\\Sigma_{\\mathbf{Y}_j}\\Y)

These marginal likelihoods are implemented relatively efficiently in
`MatnormModelBase.matnorm_logp_marginal_row` and
`MatnormModelBase.matnorm_logp_marginal_col`.

Partitioned matrix normal conditionals
--------------------------------------

Here we extend the multivariate gaussian conditional identity to matrix
normals. This is used for prediction in some models. Below, we
use lowercase subscripts for sizes to make dimensionalities easier to track.
Uppercase subscripts for covariances help keep track where they come from.


Next, we do the same for the partitioned gaussian identity. First two
vectorized matrix-normals that form our partition:

.. math::
    \\mathbf{X}_{ij} &\\sim \\mathcal{MN}(\\mathbf{A}_{ij}, \\Sigma_{i},
    \\Sigma_{j}) \\rightarrow \\vecop[\\mathbf{X}_{ij}] \\sim
    \\mathcal{N}(\\vecop[\\mathbf{A}_{ij}], \\Sigma_{j}\\otimes\\Sigma_{i})\\\\
    \\mathbf{Y}_{ik} &\\sim \\mathcal{MN}(\\mathbf{B}_{ik}, \\Sigma_{i},
    \\Sigma_{k}) \\rightarrow \\vecop[\\mathbf{Y}_{ik}] \\sim
    \\mathcal{N}(\\vecop[\\mathbf{B}_{ik}], \\Sigma_{k}\\otimes\\Sigma_{i})\\\\
    \\begin{bmatrix}\\vecop[\\mathbf{X}_{ij}] \\\\ \\vecop[\\mathbf{Y}_{ik}]
    \\end{bmatrix}
    & \\sim \\mathcal{N}\\left(\\vecop\\begin{bmatrix}\\mathbf{A}_{ij}
    \\\\ \\mathbf{B}_{ik}
    \\end{bmatrix}
    , \\begin{bmatrix} \\Sigma_{j}\\otimes \\Sigma_i  &
     \\Sigma_{jk} \\otimes \\Sigma_i  \\\\
    \\Sigma_{kj}\\otimes \\Sigma_i & \\Sigma_{k} \\otimes
     \\Sigma_i\\end{bmatrix}\\right)

We apply the standard partitioned Gaussian identity and simplify using the
properties of the :math:`\\vecop` operator and the mixed product property
of kronecker products:

.. math::
    \\vecop[\\X_{ij}] \\mid \\vecop[\\Y_{ik}]\\sim
    \\mathcal{N}(&\\vecop[\\A_{ij}] + (\\Sigma_{jk}\\otimes\\Sigma_i)
    (\\Sigma_k\\inv\\otimes\\Sigma_i\\inv)(\\vecop[\\Y_{ik}]-\\vecop[\\B_{ik}]),\\\\
    & \\Sigma_j\\otimes\\Sigma_i -  (\\Sigma_{jk}\\otimes\\Sigma_i)
    (\\Sigma_k\\inv\\otimes\\Sigma_i\\inv) (\\Sigma_{kj}\\otimes\\Sigma_i))\\\\
    =\\mathcal{N}(&\\vecop[\\A_{ij}] +
     (\\Sigma_{jk}\\Sigma_k\\inv\\otimes\\Sigma_i\\Sigma_i\\inv)
     (\\vecop[\\Y_{ik}]-\\vecop[\\B_{ik}]), \\\\
     & \\Sigma_j\\otimes\\Sigma_i -
     (\\Sigma_{jk}\\Sigma_k\\inv\\Sigma_{kj}\\otimes
     \\Sigma_i\\Sigma_i\\inv \\Sigma_i))\\\\
    =\\mathcal{N}(&\\vecop[\\A_{ij}] + (\\Sigma_{jk}\\Sigma_k\\inv\\otimes\\I)
    (\\vecop[\\Y_{ik}]-\\vecop[\\B_{ik}]), \\\\
     & \\Sigma_j\\otimes\\Sigma_i -
     (\\Sigma_{jk}\\Sigma_k\\inv\\Sigma_{kj}\\otimes\\Sigma_i)\\\\
    =\\mathcal{N}(&\\vecop[\\A_{ij}] +
    \\vecop[\\Y_{ik}-\\B_{ik}\\Sigma_k\\inv\\Sigma_{kj}],
     (\\Sigma_j-\\Sigma_{jk}\\Sigma_k\\inv\\Sigma_{kj})\\otimes\\Sigma_i)


Next, we recognize that this multivariate gaussian is equivalent to the
following matrix variate gaussian:

.. math::
    \\X_{ij} \\mid \\Y_{ik}\\sim \\mathcal{MN}(\\A_{ij} +
    (\\Y_{ik}-\\B_{ik})\\Sigma_k\\inv\\Sigma_{kj}, \\Sigma_i,
    \\Sigma_j-\\Sigma_{jk}\\Sigma_k\\inv\\Sigma_{kj})

The conditional in the other direction can be written by working through the
same algebra:

.. math::
    \\Y_{ik} \\mid \\X_{ij}\\sim \\mathcal{MN}(\\B_{ik} +(\\X_{ij}-
    \\A_{ij})\\Sigma_j\\inv\\Sigma_{jk}, \\Sigma_i,
    \\Sigma_k-\\Sigma_{kj}\\Sigma_j\\inv\\Sigma_{jk})

Finally, vertical rather than horizontal concatenation (yielding a partitioned
row rather than column covariance) can be written by recognizing the behavior
of the matrix normal under transposition:

.. math::
    \\X\\trp_{ji} \\mid \\Y\\trp_{ki}\\sim \\mathcal{MN}(&\\A\\trp_{ji} +
    \\Sigma_{jk}\\Sigma_k\\inv(\\Y\\trp_{ki}-\\B\\trp_{ki}),
     \\Sigma_j-\\Sigma_{jk}\\Sigma_k\\inv\\Sigma_{kj}, \\Sigma_i)\\\\
    \\Y\\trp_{ki} \\mid \\X\\trp_{ji}\\sim \\mathcal{MN}(&\\B\\trp_{ki} +
    \\Sigma_{kj}\\Sigma_j\\inv(\\X\\trp_{ji}-\\A\\trp_{ji}),
     \\Sigma_k-\\Sigma_{kj}\\Sigma_j\\inv\\Sigma_{jk}, \\Sigma_i)

These conditional likelihoods are implemented relatively efficiently
in `MatnormModelBase.matnorm_logp_conditional_row` and
`MatnormModelBase.matnorm_logp_conditional_col`.

"""



import tensorflow as tf
from sklearn.base import BaseEstimator
from sklearn.linear_model import LinearRegression
from .covs import CovIdentity
from brainiak.utils.utils import cov2corr
import numpy as np
from brainiak.matnormal.matnormal_likelihoods import matnorm_logp_marginal_row
from brainiak.matnormal.utils import (
    pack_trainable_vars,
    unpack_trainable_vars,
    make_val_and_grad,
    unflatten_cholesky_unique,
    flatten_cholesky_unique,
)

from scipy.optimize import minimize

__all__ = ["MNRSA"]


class MNRSA(BaseEstimator):
    """ Matrix normal version of RSA.

    The goal of this analysis is to find the covariance of the mapping from
    some design matrix X to the fMRI signal Y. It does so by marginalizing over
    the actual mapping (i.e. averaging over the uncertainty in it), which
    happens to correct a bias imposed by structure in the design matrix on the
    RSA estimate (see Cai et al., NIPS 2016).

    This implementation makes different choices about residual covariance
    relative to `brainiak.reprsimil.BRSA`: Here, the noise covariance is
    assumed to be kronecker-separable. Informally, this means that all voxels
    have the same temporal covariance, and all time points have the same
    spatial covariance. This is in contrast to BRSA, which allows different
    temporal covariance for each voxel. On the other hand, computational
    efficiencies enabled by this choice allow MNRSA to support a richer class
    of space and time covariances (anything in `brainiak.matnormal.covs`).

    For users: in general, if you are worried about voxels each having
    different temporal noise structure,you should use
    `brainiak.reprsimil.BRSA`. If you are worried about between-voxel
    correlations or temporal covaraince structures that BRSA does not
    support, you should use MNRSA.

    .. math::
        Y &\\sim \\mathcal{MN}(0, \\Sigma_t + XLL^TX^T+
        X_0X_0^T, \\Sigma_s)\\

        U &= LL^T

    Parameters
    ----------
    time_cov : subclass of CovBase
        Temporal noise covariance class following CovBase interface.
    space_cov : subclass of CovBase
        Spatial noise covariance class following CovBase interface.
    optimizer : string, Default :'L-BFGS'
        Name of scipy optimizer to use.
    optCtrl :  dict, default: None
        Additional arguments to pass to scipy.optimize.minimize.

    """

    def __init__(
        self, time_cov, space_cov, n_nureg=5, optimizer="L-BFGS-B",
        optCtrl=None
    ):

        self.n_T = time_cov.size
        self.n_V = space_cov.size
        self.n_nureg = n_nureg

        self.optMethod = optimizer
        if optCtrl is None:
            self.optCtrl = {}

        self.X_0 = tf.Variable(
            tf.random.normal([self.n_T, n_nureg], dtype=tf.float64), name="X_0"
        )

        self.train_variables = [self.X_0]

        self.time_cov = time_cov
        self.space_cov = space_cov

        self.train_variables.extend(self.time_cov.get_optimize_vars())
        self.train_variables.extend(self.space_cov.get_optimize_vars())

    @property
    def L(self):
        """
        Cholesky factor of the RSA matrix.
        """
        return unflatten_cholesky_unique(self.L_flat)

    def fit(self, X, y, naive_init=True):
        """ Estimate dimension reduction and cognitive model parameters

        Parameters
        ----------
        X: 2d array
            Brain data matrix (TRs by voxels). Y in the math
        y: 2d array or vector
            Behavior data matrix (TRs by behavioral obsevations). X in the math
        max_iter: int, default=1000
            Maximum number of iterations to run
        step: int, default=100
            Number of steps between optimizer output
        restart: bool, default=True
            If this is true, optimizer is restarted (e.g. for a new dataset).
            Otherwise optimizer will continue from where it is now (for example
            for running more iterations if the initial number was not enough).

        """

        # In the method signature we follow sklearn discriminative API
        # where brain is X and behavior is y. Internally we are
        # generative so we flip this here
        X, Y = y, X

        self.n_c = X.shape[1]

        if naive_init:
            # initialize from naive RSA
            m = LinearRegression(fit_intercept=False)
            m.fit(X=X, y=Y)
            self.naive_U_ = np.cov(m.coef_.T)
            naiveRSA_L = np.linalg.cholesky(self.naive_U_)
            self.L_flat = tf.Variable(
                flatten_cholesky_unique(naiveRSA_L), name="L_flat",
                dtype="float64"
            )
        else:
            chol_flat_size = (self.n_c * (self.n_c + 1)) // 2
            self.L_flat = tf.Variable(
                tf.random.normal([chol_flat_size], dtype="float64"),
                name="L_flat",
                dtype="float64",
            )

        self.train_variables.extend([self.L_flat])

        def lossfn(theta): return -self.logp(X, Y)
        val_and_grad = make_val_and_grad(lossfn, self.train_variables)

        x0 = pack_trainable_vars(self.train_variables)

        opt_results = minimize(fun=val_and_grad, x0=x0,
                               jac=True, method=self.optMethod, **self.optCtrl)

        unpacked_theta = unpack_trainable_vars(
            opt_results.x, self.train_variables)
        for var, val in zip(self.train_variables, unpacked_theta):
            var.assign(val)

        self.U_ = self.L.numpy().dot(self.L.numpy().T)
        self.C_ = cov2corr(self.U_)

    def logp(self, X, Y):
        """ MNRSA Log-likelihood"""

        rsa_cov = CovIdentity(size=self.n_c + self.n_nureg)
        x_stack = tf.concat([tf.matmul(X, self.L), self.X_0], 1)
        return (
            self.time_cov.logp
            + self.space_cov.logp
            + rsa_cov.logp
            + matnorm_logp_marginal_row(
                Y,
                row_cov=self.time_cov,
                col_cov=self.space_cov,
                marg=x_stack,
                marg_cov=rsa_cov,
            )
        )
